#!/usr/bin/env python
#
# update_geoip_db.py
#
# Copyright (c) 2014, DFN-CERT Services GmbH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


from urllib2 import urlopen, URLError
from gzip import GzipFile
from StringIO import StringIO
from os import rename
from ConfigParser import SafeConfigParser, NoSectionError, NoOptionError
from logging import Formatter, getLogger, StreamHandler, WARNING, INFO
from sys import stderr, argv
from logging.config import fileConfig


# default database files
DEFAULT_GEOIP_DB = '/services/inst/mixmod/data/GeoIP.dat'
DEFAULT_GEOIPV6_DB = '/services/inst/mixmod/data/GeoIPv6.dat'

# urls for geoip downloads
URL1 = 'http://geolite.maxmind.com/download/geoip/database/GeoLiteCountry/GeoIP.dat.gz'
URL2 = 'http://geolite.maxmind.com/download/geoip/database/GeoIPv6.dat.gz'

# default path to config file for production system
CONFFILE = '/services/etc/mixmod/mixmod.conf'

def download(logger, original_files):
    """
    Downloads and replaces the database files.
    """
    urls = [URL1, URL2]
    if len(original_files) != 2:
        msg = 'Two original filenames expected, %s received.' % (
            len(original_files))
        logger.critical(msg)
        exit(1)
    for url in urls:
        renamed = True
        logger.info('Downloading file %s.' % url)
        try:
            gz_data = urlopen(url)
        except URLError as e:
            msg = 'Unable to get file %s: %s' % (url, str(e))
            logger.critical(msg)
            continue
        buf = StringIO(gz_data.read())
        data = GzipFile(fileobj=buf)
        original_file = original_files.pop(0)
        logger.info('Renaming file "%s" to "%s".' % (
            original_file, original_file + '_old'))
        try:
            rename(original_file, original_file + '_old')
        except OSError as e:
            renamed = False
        logger.info('Saving downloaded database file to "%s".' % original_file)
        try:
            with open(original_file, 'w') as f:
                f.write(data.read())
        except IOError:
            msg = 'Unable to write file %s.' % original_file
            logger.critical(msg)
            if renamed:
                rename(original_file + '_old', original_file)
            continue
        finally:
            data.close()


def get_config():
    """
    Processes the configuration file and checks for options.
    """
    config = SafeConfigParser()
    if '--help' in argv:
        print 'Usage: %s [--verbose][--config FILE][--help]' % argv[0]
        exit(0)
    try:
        config_file = argv[argv.index('--config')+1]
    except ValueError:
        config_file = CONFFILE
        msg = '%s: No config file specified: Using %s.\n' % (
            argv[0], config_file)
        stderr.write(msg)
    except IndexError:
        stderr.write('Config file name missing:')
        stderr.write('Usage: %s [--verbose][--config FILE][--help]' % argv[0])
        exit(1)
    if not config.read(config_file):
        msg = '%s: Unable to load config from file. Does "%s" exist?\n' % (
            argv[0], config_file)
        stderr.write(msg)
        exit(1)
    # configuring logging
    try:
        fileConfig(config_file)
        logger = getLogger('MixModMain (from config)')
    except (NoOptionError, NoSectionError) as e:
        logger = getLogger('MixModMain (Default)')
        handler = StreamHandler()
        fmt_string = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        formatter = Formatter(fmt_string)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(WARNING)
        msg = 'Could not load config from file: %s. Using default logger.' % (
            str(e))
        logger.warning(msg)
    if '--verbose' in argv:
        logger.setLevel(INFO)
    # get GeoIP database filenames
    for section in config.sections():
        if config.has_option(section, 'GeoIP-DB-File'):
            file1 = config.get(section, 'GeoIP-DB-File')
        if config.has_option(section, 'GeoIPv6-DB-File'):
            file2 = config.get(section, 'GeoIPv6-DB-File')
    if 'file1' not in locals():
        msg = 'Could not find option "GeoIP-DB-File" in config file. '
        msg += 'Using default file %s instead.' % DEFAULT_GEOIP_DB
        logger.warning(msg)
        file1 = DEFAULT_GEOIP_DB
    if 'file2' not in locals():
        msg = 'Could not find option "GeoIPv6-DB-File" in config file. '
        msg += 'Using default file %s instead.' % DEFAULT_GEOIPV6_DB
        logger.warning(msg)
        file2 = DEFAULT_GEOIPV6_DB

    return (logger, [file1, file2])

if __name__ == '__main__':
    (logger, original_files) = get_config()
    download(logger, original_files)
