# dataprocessor.py
#
# Copyright (c) 2015  DFN-CERT Services GmbH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


class DataProcessor(object):
    """Base class for processing mixmod data.

    Data processors work a bit like queues. Data items can be added
    using the put() method, after which a varying number of processed
    data items can be taken out again. The get() method will return
    those items in turn as long as the empty() method does not return
    a true value.

    Calling the close() method tells the processor that no further
    items will be added. After this, more items can be made available
    using the get() method.

    Data processors can be used as context managers, calling close()
    when the context ends.

    """

    def __init__(self, **conf):
        self.conf = conf

    def put(self, data):
        """Add a new data object to the processor."""
        raise RuntimeError("Override the put() method in the subclass")

    def get(self):
        """Retrieve the next processed data object.

        This should return None if no items are currently available.

        """
        raise RuntimeError("Override the get() method in the subclass")

    def close(self):
        """Inform the processor that no further data items will be added.

        After this, more data items can be available.

        """
        pass

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, tb):
        self.close()

    def __iter__(self):
        obj = self.get()
        while obj is not None:
            yield obj
            obj = self.get()
