#!/usr/bin/env python
#
# main.py
#
# Copyright (c) 2015, DFN-CERT Services GmbH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

import os
import signal
import socket
from ConfigParser import NoOptionError, NoSectionError, SafeConfigParser
from datetime import datetime, timedelta
from io import FileIO
from logging import DEBUG, Formatter, StreamHandler, getLogger
from logging.config import fileConfig
from multiprocessing import Process
from multiprocessing.queues import SimpleQueue
from sys import argv

from mixmod.errors import ConfigError, DataError, FormatError, MixModError
from mixmod.config import MIXMOD_CLASSES

# default path to config file for production system
CONFFILE = '/services/etc/mixmod/mixmod.conf'


def process_config():
    """
    Processes the configuration file and checks for various options.
    """
    # Open and read config file
    config = SafeConfigParser()
    config_file = CONFFILE
    try:
        config_file = argv[argv.index('--config') + 1]
    except ValueError:
        pass
    except IndexError:
        print 'Config file name missing:'
        print 'Usage: %s file1 ... fileX [--config FILE][--help]' % argv[0]
        exit(1)
    if not config.read(config_file):
        print 'Unable to load config from file "%s". Does file "%s" exist?' % (
            config_file, config_file)
        exit(1)

    # configuring logging
    try:
        fileConfig(config_file)
        logger = getLogger('MixModMain (from config)')
    except NoSectionError:
        logger = getLogger('MixModMain (Default)')
        handler = StreamHandler()
        fmt_string = '%(asctime)s - %(name)s - %(levelname)s - %(message)s'
        formatter = Formatter(fmt_string)
        handler.setFormatter(formatter)
        logger.addHandler(handler)
        logger.setLevel(DEBUG)

    # processing config
    modul_identifier_list = list()
    src_list = list()
    configsets = dict()
    position = 1
    file_order = list()
    for section in config.sections():
        if section[0:7] == 'FlowMod':
            dict_values = dict(config.items(section))
            dict_values['name'] = section[8:]
            dict_values['confsection'] = dict_values['name']
            modul_identifier_list.append(dict_values['name'])
            try:
                dict_values['confsection'] = dict_values['config']
            except KeyError:
                dict_values['confsection'] = dict_values['name']
            dict_values['config'] = config
            try:
                dict_values['class_text'] = dict_values['class']
            except KeyError:
                msg = 'Option "class" in section %s in configfile missing.' % (
                    dict_values['name'])
                logger.critical(msg)
                exit(1)
            try:
                dict_values['class'] = MIXMOD_CLASSES[dict_values['class']]
            except KeyError:
                msg = 'Unknown class "%s" in section %s of config file.' % (
                    dict_values['class'], dict_values['name'])
                logger.critical(msg)
                exit(1)
            if 'Import' in dict_values['class_text']:
                if 'listen' in dict_values:
                    try:
                        dict_values['host'], dict_values['port'] = dict_values['listen'].split(':')
                        dict_values['port'] = int(dict_values['port'])
                    except ValueError:
                        logger.critical('Cannot find host:port in config item "{}"'.format(
                            dict_values['listen']))
                        exit(1)
                    dict_values['func'] = UDPImporter
                    dict_values['src'] = ''
                else:
                    try:
                        file_order.append((position, dict_values['class_text'][6:],
                                           'Importer', argv[position]))
                    except IndexError:
                        file_order.append((position, dict_values['class_text'][6:], 'Importer'))
                    dict_values['func'] = Importer
                    dict_values['position'] = position
                    try:
                        dict_values['filename'] = argv[position]
                    except IndexError:
                        pass
                    dict_values['src'] = ''
                    position += 1
            elif 'Export' in dict_values['class_text']:
                try:
                    file_order.append((position, dict_values['class_text'][6:],
                                       'Exporter', argv[position]))
                except IndexError:
                    file_order.append((position, dict_values['class_text'][6:],
                                       'Exporter'))
                dict_values['func'] = Exporter
                dict_values['position'] = position
                try:
                    dict_values['filename'] = argv[position]
                except IndexError:
                    pass
                position += 1
            elif 'Pseudo' in dict_values['class_text']:
                dict_values['func'] = Intermediate
            elif 'Filter' in dict_values['class_text']:
                dict_values['func'] = Intermediate
            elif 'Aggregator' in dict_values['class_text']:
                dict_values['func'] = Intermediate
            elif 'Sorter' in dict_values['class_text']:
                dict_values['func'] = Intermediate
            configsets[dict_values['name']] = dict_values

    # calculating dst tuple
    for name, dict_values in configsets.items():
        dict_values['dst'] = ()
        for key, values in configsets.items():
            if 'src' in values:
                liste = [x.strip() for x in values['src'].split(',')]
                if name in liste:
                    dict_values['dst'] = dict_values['dst'] + (key,)

    # calculating src tuple
    for name, dict_values in configsets.items():
        src_tuple = ()
        liste = [x.strip() for x in dict_values['src'].split(',')]
        if '' in liste:
            dict_values['src'] = ()
        else:
            for elem in liste:
                src_tuple = src_tuple + (elem,)
                src_list.append(elem)
            dict_values['src'] = src_tuple

    # is src an existing module identifier
    for elem in src_list:
        if elem not in modul_identifier_list:
            msg = 'Error in configuration: %s is not a correct module ' \
                'identifier.' % elem
            logger.critical(msg)
            exit(1)

    # checking for errors in command line input
    if '--config' in argv:
        position += 2
    if len(argv) < 2 or '--help' in argv or position != len(argv):
        print 'Usage: %s' % argv[0],
        for i, f in enumerate(file_order):
            print 'FILE' + str(i + 1),
        print '[--config FILE][--help]'
        for f in file_order:
            print '\tFILE %s: %s file (%s)' % (f[0], f[1], f[2])
        exit(1)
    else:
        for f in file_order:
            print 'File %s: %s file (%s): %s' % f
    return logger, configsets


class MixModProcess(Process):
    def __init__(self, config, queue, src, dst, logger):
        super(MixModProcess, self).__init__()
        self.name = config['name']
        self.config = config
        self.configparser = config.get('config')
        self.confsection = config.get('confsection')
        self.class_type = config['class']
        self.queue = queue
        self.src = src
        self.dst = dst
        self.logger = logger
        self.src_done_count = 0

    def get_class_config(self):
        if self.configparser is not None and self.configparser.has_section(self.confsection):
            return dict(self.configparser.items(self.confsection))
        else:
            return {}

    def send(self, data):
        for dst in self.dst:
            dst.put(data)

    def receive(self):
        while True:
            data = self.queue.get()
            if data is None:
                self.src_done_count += 1
                if self.src_done_count == len(self.src):
                    return
            else:
                yield data

    def finish(self):
        for dst in self.dst:
            dst.put(None)


class Importer(MixModProcess):
    """
    Process of the file importer.
    """
    def __init__(self, *arg, **kwargs):
        super(Importer, self).__init__(*arg, **kwargs)
        self.filein = self.config['filename']

    def run(self):
        self.logger.debug('MixModProcess %s (PID %s) start' % (self.name, os.getpid()))
        signal.signal(signal.SIGINT, lambda signum, frame: self.finish_and_exit())
        signal.signal(signal.SIGTERM, lambda signum, frame: self.finish_and_exit())

        try:
            with FileIO(self.filein, mode="r") as f:
                importer = self.class_type(f, **self.get_class_config())
                for data in importer:
                    self.send(data)
        except IOError as e:
            msg = '%s: Unable to open and read file %s: %s' % (
                self.name, self.filein, str(e))
            self.logger.critical(msg)
            exit(1)
        except FormatError as e:
            self.logger.critical(self.name + ': ' + str(e))
            exit(1)
        finally:
            self.finish()
            self.logger.debug('MixModProcess %s finish' % self.name)

    def finish_and_exit(self):
        self.finish()
        exit(1)


class UDPImporter(MixModProcess):
    """
    Importer opening, binding to, and reading data from a UDP socket instead of a file
    """
    def __init__(self, *args, **kwargs):
        super(UDPImporter, self).__init__(*args, **kwargs)
        self.host = self.config['host']
        self.port = self.config['port']
        self.socket = None
        try:
            self.udp_rxbufsize = int(self.config['udp_rxbufsize'])
        except (KeyError, ValueError):
            self.udp_rxbufsize = 65535

    def run(self):
        self.logger.debug('MixModProcess %s (PID %s) start' % (self.name, os.getpid()))
        signal.signal(signal.SIGINT, lambda signum, frame: self.finish_and_exit())
        signal.signal(signal.SIGTERM, lambda signum, frame: self.finish_and_exit())

        fobj = self.start_udp_receiver()
        error_reported = False

        while True:
            try:
                importer = self.class_type(fobj, **self.get_class_config())
                for mm_data in importer:
                    if error_reported:
                        self.logger.info(self.name + ": Importer working again")
                        error_reported = False
                    self.send(mm_data)
            except socket.error as e:
                self.logger.critical(self.name + ': ' + str(e))
                break
            except Exception as e:
                if not error_reported:
                    self.logger.critical(self.name + ': ' + str(e))
                    error_reported = True
                fobj.reset()

    def start_udp_receiver(self):
        msg = 'unknown error'
        for gai in socket.getaddrinfo(self.host, self.port, socket.AF_UNSPEC, socket.SOCK_DGRAM, 0):
            af, stype, prot, cname, saddr = gai
            try:
                self.socket = socket.socket(af, stype, prot)
                self.socket.setsockopt(socket.SOL_SOCKET, socket.SO_REUSEADDR, 1)
            except socket.error as msg:
                self.socket = None
                continue
            try:
                self.socket.bind(saddr)
            except socket.error as msg:
                self.socket.close()
                self.socket = None
                continue
            else:
                break
        if self.socket is None:
            self.logger.error('could not listen on {}:{}/udp: {}'.format(
                self.host, self.port, msg))

        self.logger.debug('waiting for datagrams on {}:{}/udp'.format(self.host, self.port))
        return SocketFile(self.socket)

    def finish_and_exit(self):
        self.finish()
        exit(1)


class SocketFile(object):
    def __init__(self, s, readsize=65535):
        self.socket = s
        self.buf = ""
        self.readsize = readsize

    def read(self, size):
        while len(self.buf) < size:
            data, addr = self.socket.recvfrom(self.readsize)
            self.buf += data
        data = self.buf[0:size]
        self.buf = self.buf[size:]
        return data

    def reset(self):
        self.buf = ""


class Intermediate(MixModProcess):
    """
    MixModProcess of on intermediate module like filter or pseudonymizer.
    """
    def __init__(self, *args, **kwargs):
        super(Intermediate, self).__init__(*args, **kwargs)
        self.src_done_count = 0

    def run(self):
        self.logger.debug('MixModProcess %s (PID %s) start' % (self.name, os.getpid()))
        # We ignore these signals as we receive None from the
        # importers for them.
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        signal.signal(signal.SIGTERM, signal.SIG_IGN)
        try:
            intermediate = self.class_type(**self.get_class_config())
        except (EnvironmentError, ValueError, ConfigError, NoOptionError) as e:
            self.logger.critical(self.name + ': ' + str(e))
            self.finish()
            exit(1)

        try:
            with intermediate as proc:
                for data_in in self.receive():
                    proc.put(data_in)
                    for data_out in proc:
                        self.send(data_out)
            for data_out in proc:
                self.send(data_out)
        except DataError as e:
            self.logger.critical(self.name + ": " + str(e))
        except Exception as e:
            self.logger.exception(self.name + ": " + str(e))
        finally:
            self.finish()
            self.logger.debug('MixModProcess %s finish' % self.name)


class Exporter(MixModProcess):
    """
    MixModProcess of the file exporter.
    """
    def __init__(self, *args, **kwargs):
        super(Exporter, self).__init__(*args, **kwargs)
        self.fileout = self.config['filename']
        try:
            self.timeslice_seconds = int(self.config['timeslice'])
        except (KeyError, ValueError):
            self.timeslice_seconds = None
            self.timeslice_timedelta = None
        else:
            self.timeslice_timedelta = timedelta(seconds=self.timeslice_seconds)

    def run(self):
        self.logger.debug('MixModProcess %s (PID %s) start' % (self.name, os.getpid()))
        # We ignore these signals as we receive None from the
        # importers for them.
        signal.signal(signal.SIGINT, signal.SIG_IGN)
        signal.signal(signal.SIGTERM, signal.SIG_IGN)

        if self.timeslice_timedelta is not None:
            starttime = datetime.now()
            endtime = starttime + self.timeslice_timedelta
        else:
            starttime = None
            endtime = None

        outfile = FileIO(self.get_file_name(starttime), mode="w")
        exporter = self.class_type(outfile, **self.get_class_config())
        try:
            for data in self.receive():
                if endtime is not None and datetime.now() >= endtime:
                    try:
                        exporter.close()
                    except MixModError as e:
                        self.logger.critical("Error {} in exporter {}: {}"
                                             .format(type(e).__name__,
                                                     self.class_type.__name__,
                                                     e))
                    except Exception as e:
                        self.logger.exception("Exception in export: {}".format(e))
                    outfile.close()
                    starttime = endtime
                    endtime = starttime + self.timeslice_timedelta
                    outfile = FileIO(self.get_file_name(starttime), mode="w")
                    exporter = self.class_type(outfile, **self.get_class_config())
                exporter.put(data)
        except MixModError as e:
            self.logger.critical("Error {} in exporter {}: {}"
                                 .format(type(e).__name__,
                                         self.class_type.__name__,
                                         e))
        except Exception as e:
            self.logger.exception("Exception in export: {}".format(e))
        finally:
            try:
                exporter.close()
            except Exception as e:
                self.logger.critical("Error {} in exporter {}: {}"
                                     .format(type(e).__name__,
                                             self.class_type.__name__,
                                             e))
            try:
                outfile.close()
            except:
                pass

        self.logger.debug('MixModProcess %s finish' % self.name)

    def get_file_name(self, timestamp):
        if timestamp is None:
            return self.fileout
        else:
            return '{}-{}'.format(
                self.fileout,
                timestamp.strftime('%Y%m%d-%H%M%S')
            )


def main():
    """
    Main function of the script.
    """
    processes = dict()
    queues = dict()
    (logger, configs) = process_config()

    # create queues
    for name, config in configs.items():
        queues[name] = SimpleQueue()

    # We ignore these signals, as they're handled in the importer
    # processes and we just terminate anyhow.
    signal.signal(signal.SIGINT, signal.SIG_IGN)
    signal.signal(signal.SIGTERM, signal.SIG_IGN)

    # create processes
    for name, config in configs.items():
        dst = list()
        src = list()
        for d in config['dst']:
            try:
                dst.append(queues[d])
            except KeyError as e:
                msg = 'Error in configuration: % does not exist.' % str(e)
                logger.critical(msg)
                exit(1)
        for s in config['src']:
            try:
                src.append(queues[s])
            except KeyError as e:
                msg = 'Error in configuration: %s does not exist.' % str(e)
                logger.critical(msg)
                exit(1)
        processes[name] = config['func'](config, queues[name], src, dst, logger)
        processes[name].start()


if __name__ == '__main__':
    main()
