# netflow9importer.py
#
# Copyright (c) 2015 DFN-CERT Services GmbH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


from datetime import datetime, timedelta
import struct

from netaddr import IPAddress

from mixmod import importer
from mixmod.errors import FormatError

# setting endian type
ENDIAN_TYPE = '>'  # > big endian
# < little endian

# setting relevant netflow/ipfix fields
MIXMOD_SPEC = {
    1: 'byte_count',  # = IN_BYTES
    2: 'packet_count',  # = IN_PKTS
    4: 'ip_protocol_number',
    5: 'type_of_service',  # = SRC_TOS
    6: 'tcp_flags',
    7: 'src_port',
    8: 'src_ip_v4',
    10: 'snmp_index_in',
    11: 'dst_port',
    12: 'dst_ip_v4',
    14: 'snmp_index_out',
    16: 'src_as',
    17: 'dst_as',
    21: 'utc_last_packet',
    22: 'utc_first_packet',
    # 23: 'byte_count_out',
    # 24: 'packet_count_out',
    27: 'src_ip_v6',
    28: 'dst_ip_v6',
    # 55: 'type_of_service_dst',
    60: 'ip_version',
    61: 'direction',
    152: 'utc_first_packet',
    153: 'utc_last_packet'}

# setting length formats
LENGTH_TO_FORMAT_CHAR = {
    1: struct.Struct('>B'),
    2: struct.Struct('>H'),
    4: struct.Struct('>I'),
    8: struct.Struct('>Q'),
    16: struct.Struct('>QQ')}

# Types of fields that are representing an IP address
IP_ADDR = {8, 12, 15, 18, 27, 28, 62, 63}

# Types of fields that are representing time and dates
RELATIVE_DATETIMES = {21, 22}
ABSOLUTE_DATETIMES = {152, 153}


# definition of netflow field order and length
HEADER_FORMAT = struct.Struct(
    ENDIAN_TYPE +
    'H'  # uint16 Version
    'H'  # uint16 Count
    'I'  # uint32 System Uptime
    'I'  # uint32 Unix Seconds
    'I'  # uint32 Package Sequence
    'I')  # uint32 Source ID

MANGLED_HEADER_FORMAT = struct.Struct(
    ENDIAN_TYPE +
    'I'  # uint32 System Uptime
    'I'  # uint32 Unix Seconds
    'I'  # uint32 Package Sequence
    'I')  # uint32 Source ID

# definition of flowset type format string
FLOWSET_TYPE_FORMAT = struct.Struct(
    ENDIAN_TYPE +
    'H'  # uint16 Flowset ID
    'H')  # uint16 Length in Bytes
FLOWSET_ID_TEMPLATE = 0
FLOWSET_ID_OPTION = 1
FLOWSET_ID_DATA = 255

# definition of template header format string
TEMPLATE_HEADER_FORMAT = struct.Struct(
    ENDIAN_TYPE +
    'H'  # uint16 Template ID
    'H')  # uint16 Field Count

# definition of template field format string
TEMPLATE_FIELD_FORMAT = struct.Struct(
    ENDIAN_TYPE +
    'H'  # uint16 Field Type
    'H')  # uint16 Field Length


class Netflow9Importer(importer.FileImporter):
    """Import netflow version 9 data to mixmod."""
    DIRECTION_INGRESS = 0
    DIRECTION_EGRESS = 1

    CONFIG = {
        "forced_sys_start_time": {
            "type": "/^\\d\\d\\d\\d-\\d\\d-\\d\\dT\\d\\d:\\d\\d:\\d\\d$/",
            "default": None
        },
        "standard_direction": {
            "type": "int",
            "default": DIRECTION_INGRESS
        }
    }

    def __init__(self, fobj,
                 forced_sys_start_time=None,
                 standard_direction=DIRECTION_INGRESS,
                 **conf):
        super(Netflow9Importer, self).__init__(fobj, **conf)
        self.templates = dict()
        self.standard_direction = int(standard_direction)
        if isinstance(forced_sys_start_time, basestring):
            self.forced_sys_start_time = datetime.strptime(forced_sys_start_time,
                                                           '%Y-%m-%dT%H:%M:%S')
        else:
            self.forced_sys_start_time = forced_sys_start_time

        self.iterator = self._parse()

    def get(self):
        try:
            return next(self.iterator)
        except StopIteration:
            return None

    def _parse(self):
        """Parses the data by using helper functions and generates mixmoddata
        dictionaries.

        """
        last_packet_was_mangled = False
        flowset_id = None
        length = None
        while True:
            if not last_packet_was_mangled:
                start_time, count = self._read_packet_header()
            else:
                start_time, count = self._read_packet_header(flowset_id, length)
                last_packet_was_mangled = False
            for i in xrange(count):
                flowset_header_length = FLOWSET_TYPE_FORMAT.size
                flowset_id, length = self.read_struct(FLOWSET_TYPE_FORMAT)
                if flowset_id == FLOWSET_ID_TEMPLATE:
                    data = self.fobj.read(length - flowset_header_length)
                    self._parse_template_flowset(data)
                elif flowset_id == FLOWSET_ID_OPTION:
                    raise FormatError("Unsupported flowset ID {} (Options)"
                                      .format(FLOWSET_ID_OPTION))
                elif flowset_id > FLOWSET_ID_DATA:
                    data = self.fobj.read(length - flowset_header_length)
                    for mixmoddata in self._parse_data_flowset(data, flowset_id, start_time):
                        yield mixmoddata
                elif flowset_id == 9:
                    # For some reason, pmacct terminates packets
                    # before they're finished and just sends a new
                    # packet. The flowset_id is the version field and
                    # the length is the count field of the new header.
                    last_packet_was_mangled = True
                    break
                else:
                    msg = 'Undefined Flowset received with Flowset ID %s.'
                    raise FormatError(msg % (flowset_id,))

    def _read_packet_header(self, override_version=None, override_count=None):
        if override_count is None:
            (version, count, system_uptime, unix_seconds,
             package_sequence, source_id) = self.read_struct(HEADER_FORMAT)
        else:
            version, count = override_version, override_count
            (system_uptime, unix_seconds,
             package_sequence, source_id) = self.read_struct(MANGLED_HEADER_FORMAT)

        if version != 9:
            raise FormatError("Unsupported Netflow version: {}"
                              .format(version))
        if self.forced_sys_start_time:
            start_time = self.forced_sys_start_time
        else:
            start_time = (datetime.utcfromtimestamp(unix_seconds) -
                          timedelta(milliseconds=system_uptime))
        return start_time, count

    def _parse_template_flowset(self, data):
        """Parses the template flowset header and writes the data format
        definitions in the list self.templates as a touple (element
        id, length in bytes). Takes the length of the flowset as an
        argument.

        """
        offset = 0
        (template_id, field_count) = TEMPLATE_HEADER_FORMAT.unpack_from(data, offset)
        offset += TEMPLATE_HEADER_FORMAT.size
        if template_id <= FLOWSET_ID_DATA:
            msg = 'Template ID %s invalid.' % (template_id,)
            raise FormatError(msg)
        else:
            self.templates[template_id] = list()
        for i in xrange(field_count):
            (field_type, field_length) = TEMPLATE_FIELD_FORMAT.unpack_from(data, offset)
            offset += TEMPLATE_FIELD_FORMAT.size
            self.templates[template_id].append((field_type, field_length))

    def _parse_data_flowset(self, data, template_id, start_time):
        """
        Parses a data flowset and generates mixmoddata dictionaries. It takes
        the length of the flowset and the template ID to be used as an
        argument. A flowset can contain several records.
        """
        offset = 0
        if template_id not in self.templates:
            msg = 'Unknown Template ID "{}", have templates: {}.'.format(
                template_id, self.templates.keys()
            )
            raise FormatError(msg)

        single_data_size = sum(field_type_length[1]
                               for field_type_length in self.templates[template_id])
        while offset + single_data_size <= len(data):
            mixmoddata = {'original_format': 'nf9',
                          'direction': self.standard_direction}

            for field_type_length in self.templates[template_id]:
                try:
                    field_format = LENGTH_TO_FORMAT_CHAR[field_type_length[1]]
                except KeyError:
                    msg = 'Unknown field length %s.' % (field_type_length[1],)
                    raise FormatError(msg)

                values = field_format.unpack_from(data, offset)
                offset += field_format.size
                # special threatment for 128bit integer
                if len(values) == 2:
                    (field1, field2) = values
                    field = field1 * 18446744073709551616L + field2
                else:
                    (field,) = values

                # storing field number or name and value in mixmoddata dict
                if field_type_length[0] in IP_ADDR:
                    tmp = IPAddress(field)
                elif field_type_length[0] in RELATIVE_DATETIMES:
                    tmp = start_time + timedelta(milliseconds=field)
                elif field_type_length[0] in ABSOLUTE_DATETIMES:
                    tmp = datetime.utcfromtimestamp(field / 1000.0)
                else:
                    tmp = field
                if field_type_length[0] in MIXMOD_SPEC:
                    mixmoddata[MIXMOD_SPEC[field_type_length[0]]] = tmp
                else:
                    mixmoddata[field_type_length[0]] = tmp
            yield mixmoddata

    def read_struct(self, fmt):
        data = self.fobj.read(fmt.size)
        try:
            return fmt.unpack(data)
        except struct.error as e:
            raise FormatError("Unexpected end of file in data: {}"
                              .format(e))
