# nfcapimporter.py
#
# Copyright (c) 2014, DFN-CERT Services GmbH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


from datetime import datetime, timedelta

import netaddr

from mixmod import importer
from mixmod import nfcap

# List of fields directly transferred from nfcap records to mixmod
# data objects.
FIELD_LIST = ('fwd_status', 'src_port', 'dst_port', 'src_as',
              'dst_as', 'src_tos', 'tos', 'dir', 'srcmask',
              'dstmask', 'next_hop_ip', 'next_hop_ip_version',
              'next_bgp_ip', 'next_bgp_ip_version', 'srcvlan',
              'dstvlan', 'numoutpackets', 'numoutbyts',
              'aggrflows', 'in_src_mac', 'out_dst_mac',
              'in_dst_mac', 'out_src_mac', 'mpls_label_1',
              'mpls_label_2', 'mpls_label_3', 'mpls_label_4',
              'mpls_label_5', 'mpls_label_6', 'mpls_label_7',
              'mpls_label_8', 'mpls_label_9', 'mpls_label_10',
              'router_ip', 'router_ip_version', 'fill_e25',
              'engine_type', 'engine_id')


class NFCapImporter(importer.FileImporter):
    CONFIG = {}

    def __init__(self, fobj, **conf):
        super(NFCapImporter, self).__init__(fobj, **conf)
        self.nf = nfcap.NFCap(fobj.read())
        self.records = self.nf.iter_records()

    def get(self):
        try:
            return self._translate_common_record(next(self.records))
        except StopIteration:
            return None

    def _translate_common_record(self, rec):
        data = {}
        for element in FIELD_LIST:
            if hasattr(rec, element):
                data[element] = getattr(rec, element)
            else:
                data[element] = None

        data['ident'] = self.nf.ident
        data['original_format'] = 'nfc'
        data['utc_first_packet'] = (datetime.utcfromtimestamp(rec.first) +
                                    timedelta(milliseconds=rec.msec_first))
        data['utc_last_packet'] = (datetime.utcfromtimestamp(rec.last) +
                                   timedelta(milliseconds=rec.msec_last))
        data['tcp_flags'] = rec.tcp_flags
        data['ip_protocol_number'] = rec.prot
        data['type_of_service'] = rec.src_tos
        data['snmp_index_in'] = rec.input
        data['snmp_index_out'] = rec.output

        if data['next_hop_ip_version'] in (4, 6):
            data['next_hop_ip'] = netaddr.IPAddress(data['next_hop_ip'])

        if data['router_ip_version'] in (4, 6):
            data['router_ip'] = netaddr.IPAddress(data['router_ip'])

        if data['next_bgp_ip_version'] in (4, 6):
            data['next_bgp_ip'] = netaddr.IPAddress(data['next_bgp_ip'])

        if data['src_as'] == 0:
            del data['src_as']

        if data['dst_as'] == 0:
            del data['dst_as']

        if rec.ipversion == 4:
            data['ip_version'] = 4
            data['src_ip_v4'] = netaddr.IPAddress(rec.srcip)
            data['dst_ip_v4'] = netaddr.IPAddress(rec.dstip)
        elif rec.ipversion == 6:
            data["ip_version"] = 6
            data["src_ip_v6"] = netaddr.IPAddress(rec.srcip)
            data["dst_ip_v6"] = netaddr.IPAddress(rec.dstip)

        if rec.prot == nfcap.IPPROTO_ICMP:
            data['icmp_type_ipv4'] = rec.icmptype
            data['icmp_code_ipv4'] = rec.icmpcode

        if rec.prot == nfcap.IPPROTO_ICMPV6:
            data['icmp_type_ipv6'] = rec.icmptype
            data['icmp_code_ipv6'] = rec.icmpcode

        data["packet_count"] = rec.numpackets
        data["byte_count"] = rec.numbytes

        return data
