# pseudonymizer.py
#
# Copyright (c) 2014, 2015 DFN-CERT Services GmbH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from collections import deque
import hashlib
import random

from netaddr import IPAddress
try:
    import aapi
except ImportError:
    class AAPIWrapper(object):
        def __getattr__(self, name):
            raise EnvironmentError("The AAPI library is not installed")

    aapi = AAPIWrapper()

from mixmod.errors import ConfigError, DataError
from mixmod import dataprocessor


class Pseudonymizer(dataprocessor.DataProcessor):
    """This class implements anonymization and pseudonymization for MixModData.

    Anonymization means that data is replaced by random values.

    Pseudonymization means that the same original value is replaced by
    the same pseudonymous value every time it is found to enable
    identity checks, but without the ability to calculate the
    original.

    """

    MODE_ANON = 1
    MODE_PSEUDO = 2
    # When one of these fields is updated, also set the assigned mode
    # field.
    MODE_FIELDS = {
        "dst_ip_v4": "dst_mode",
        "dst_ip_v6": "dst_mode",
        "src_ip_v4": "src_mode",
        "src_ip_v6": "src_mode",
        "c2_ip_v4": "c2_mode",
        "c2_ip_v6": "c2_mode",
    }

    CONFIG = {}

    def __init__(self, **conf):
        """
        This method initializes a Pseudonymize instance.

        @type  config: ConfigParser.SafeConfigParser
        @param config: A ConfigParser instance into which the MixMod config
                       file has been read

        @type  confsection: string
        @param confsection: An alternate configuration section name to use
        """
        self.field_mapper = {}

        if "@aes_key_set" in conf:
            aapi.aes_key_set(conf["@aes_key_set"])
        self.domain_seed = conf.get("@domain_seed")

        for option, value in conf.items():
            if option.startswith("@"):
                continue
            field = option.lower()
            action, args, kwargs = _parse_action(value)

            mapper = getattr(self, "action_{}".format(action), None)
            if mapper is None:
                raise EnvironmentError('Invalid action for {}: {}'
                                       .format(field, action))

            self.field_mapper[field] = (mapper, args, kwargs)

        self.results = deque()

    def put(self, datadict):
        """The public method for pseudonymizing a data set

        @type  datadict: MixMod dataset dictionary
        @param datadict: Input data set

        @rtype: Mixmod dataset dictionary
        @return: The input dictionary with the configured replacements done
        """
        datadict = datadict.copy()
        for field in datadict.keys():
            mapper = self.field_mapper.get(field)
            if mapper is not None:
                func, args, kwargs = mapper
                func(datadict, field, *args, **kwargs)
        self.results.append(datadict)

    def get(self):
        try:
            return self.results.popleft()
        except IndexError:
            return None

    def update_mode(self, datadict, field, new_mode):
        mode_field = self.MODE_FIELDS.get(field)
        if mode_field is None:
            return
        if (
                new_mode == self.MODE_PSEUDO and
                datadict.get(mode_field) == self.MODE_ANON
        ):
            # Do not update anonymous to pseudonymous
            return
        datadict[mode_field] = new_mode

    #################
    # Action handlers

    def action_ip_aes_hash(self, datadict, field):
        """Map an IP address to an AES hash.

        This uses the @aes_key_set option, and does B{not preserve
        prefix information}.

        """
        old_ip = _get_ip_field(datadict, field)

        if old_ip.version == 4:
            new_ip = IPAddress(aapi.ip_aes_hash(int(old_ip)), 4)
        else:
            result_words = aapi.ip6_aes_hash(old_ip.words)
            out = 0
            for n in range(8):
                out += (result_words[n] << ((7 - n) * 16))
            new_ip = IPAddress(out, 6)

        datadict[field] = new_ip
        self.update_mode(datadict, field, self.MODE_PSEUDO)

    action_ip6_aes_hash = action_ip_aes_hash

    def action_ip_map(self, datadict, field):
        """Map an IP address to a dynamically growing mapping.

        Mappings change between processes.

        This method does B{not preserve prefix information}.

        """
        old_ip = _get_ip_field(datadict, field)

        if old_ip.version == 4:
            new_ip = IPAddress(aapi.ip_map(int(old_ip)), 4)
        else:
            result_words = aapi.ip6_map(old_ip.words)
            out = 0
            for n in range(8):
                out += (result_words[n] << ((7 - n) * 16))
            new_ip = IPAddress(out, 6)

        datadict[field] = new_ip
        self.update_mode(datadict, field, self.MODE_PSEUDO)

    action_ip6_map = action_ip_map

    def action_prefix_anon(self, datadict, field):
        '''
        Return an IPv4 address according to a dynamically grown mapping table.
        This method works prefix-preserving.
        '''
        old_ip = _get_ip_field(datadict, field)

        if old_ip.version == 4:
            new_ip = IPAddress(aapi.ip_pref_pres_map(int(old_ip)), 4)
        else:
            raise ConfigError("Can't apply prefix_anon on IPv6 field: {}"
                              .format(field))

        datadict[field] = new_ip
        self.update_mode(datadict, field, self.MODE_PSEUDO)

    def action_prefix_hash(self, datadict, field):
        '''
        Return an AES-based hash for the given IPv4 address, using the
        configured key.
        This method works prefix-preserving.
        '''
        old_ip = _get_ip_field(datadict, field)

        if old_ip.version == 4:
            new_ip = IPAddress(aapi.ip_pref_pres_anon(int(old_ip)), 4)
        else:
            raise ConfigError("Can't apply prefix_hash on IPv6 field: {}"
                              .format(field))

        datadict[field] = new_ip
        self.update_mode(datadict, field, self.MODE_PSEUDO)

    def action_ip_random(self, datadict, field, hostbits=8):
        '''
        Replaces the lower C{hostbits} bit of an IPv4 address with a random
        value.
        '''
        old_ip = _get_ip_field(datadict, field)
        hostbits = int(hostbits)

        if old_ip.version == 4:
            new_ip = IPAddress(aapi.ip_random(int(old_ip), hostbits), 4)
        else:
            raise ConfigError("Can't apply ip_random on IPv6 field: {}"
                              .format(field))

        datadict[field] = new_ip
        self.update_mode(datadict, field, self.MODE_ANON)

    def action_ip_truncate(self, datadict, field, hostbits=8):
        '''
        Truncates the lower C{hostbits} bit of an IPv4 address, setting them
        to 0.
        '''
        old_ip = _get_ip_field(datadict, field)
        hostbits = int(hostbits)

        if old_ip.version == 4:
            new_ip = IPAddress(aapi.ip_truncate(int(old_ip), hostbits), 4)
        else:
            raise ConfigError("Can't apply ip_truncate on IPv6 field: {}"
                              .format(field))

        datadict[field] = new_ip
        self.update_mode(datadict, field, self.MODE_ANON)

    def action_ip6_truncate_high(self, datadict, field, masklen=64):
        '''
        Sets the upper C{masklen} bit of the given IPv6 address to 0.
        '''
        old_ip = _get_ip_field(datadict, field)
        masklen = int(masklen)

        if old_ip.version == 4:
            raise ConfigError("Can't apply ip6_truncate_high on IPv4 field: {}"
                              .format(field))
        else:
            mask = _v6mask_high(masklen)
            addr = old_ip.words
            out = 0
            for n in range(8):
                masked = (addr[n] & mask[n])
                out += (masked << ((7 - n) * 16))
            new_ip = IPAddress(out, 6)

        datadict[field] = new_ip
        self.update_mode(datadict, field, self.MODE_ANON)

    def action_ip6_truncate_low(self, datadict, field, masklen=64):
        '''
        Sets the lower C{masklen} bit of the given IPv6 address to 0.
        '''
        old_ip = _get_ip_field(datadict, field)
        masklen = int(masklen)

        if old_ip.version == 4:
            raise ConfigError("Can't apply ip6_truncate_high on IPv4 field: {}"
                              .format(field))
        else:
            masklen = int(masklen)
            mask = _v6mask_low(masklen)
            addr = old_ip.words
            out = 0
            for n in range(8):
                masked = (addr[n] & mask[n])
                out += (masked << ((7 - n) * 16))
            new_ip = IPAddress(out, 6)

        datadict[field] = new_ip
        self.update_mode(datadict, field, self.MODE_ANON)

    def action_domain_anon(self, datadict, field, plain=0, maxparts=0):
        '''Anonymize domain names.

        Anonymous means that the domain is replaced by a random string
        without the ability to translate it back to the original, nor
        any requirements that the same domain should map to the same
        translated value.

        The algorithm will retain the last C{plain} parts of the
        domain name in its original form. It will also result in at
        most C{maxparts} parts, so that all parts beyond those
        replaced by a single element.

        '''
        _translate_domain(
            "domain_anon", datadict, field, plain, maxparts,
            lambda parts: _random_domain_part())

    def action_domain_pseudo(self, datadict, field, plain=0, maxparts=0):
        '''Pseudonymize domain names.

        Pseudonymous means that the domain is replaced by a string
        without the ability to translate it back to the original. But
        the same domain will map to the same translated value.

        The algorithm will retain the last C{plain} parts of the
        domain name in its original form. It will also result in at
        most C{maxparts} parts, so that all parts beyond those
        replaced by a single element.

        '''
        _translate_domain(
            "domain_anon", datadict, field, plain, maxparts,
            lambda parts: _pseudonymous_domain_part(parts, self.domain_seed))


def _parse_action(value):
    """Parse an action argument to a config option.

    Config options are formatted like this:

    field: action arg1 arg2 key1=arg3 key2=arg4

    This function will take the value there and return a tuple with
    three elements:

    ("action", ["arg1", "arg2"], {"key1": "arg3", "key2": "arg4"})

    """
    args = []
    kwargs = {}
    for option in value.split():
        if "=" in option:
            key, value = option.split("=", 1)
            kwargs[key] = value
        else:
            args.append(option)

    return args[0], args[1:], kwargs


def _get_ip_field(datadict, field):
    if field not in datadict:
        return None
    value = datadict[field]
    if not isinstance(value, IPAddress):
        try:
            value = IPAddress(value)
        except Exception:
            raise DataError("Field is not an IPAddress: {}"
                            .format(field))
    return value


def _v6mask_high(n):
    '''
    Fills 8 16-bit words with n 1s, starting from the LSB

    @type  n: integer
    @param n: Bits to be set high in return value

    @rtype: uint16_t[8]
    @return: List of intergers to be used as IPv6 bitmask
    '''
    maskwords = []
    n = 128 - n
    for i in range(8, 0, -1):
        upper = 16 * i
        lower = 16 * (i - 1) + 1
        if (n < lower):
            maskwords.append(0)
        elif (n > upper):
            maskwords.append(0xffff)
        else:
            maskwords.append(int((2 ** ((n + 1) - lower)) - 1))
    return maskwords


def _v6mask_low(n):
    '''
    fills 8 16-bit words with n 1s, starting from the MSB

    @type  n: integer
    @param n: Bits to be set high in return value

    @rtype: uint16_t[8]
    @return: List of intergers to be used as IPv6 bitmask
    '''
    maskwords = []
    for i in range(8, 0, -1):
        upper = 16 * i
        lower = 16 * (i - 1) + 1
        if (n < lower):
            maskwords.append(0xffff)
        elif (n > upper):
            maskwords.append(0)
        else:
            maskword = int((2 ** ((n + 1) - lower)) - 1)
            maskwords.append(maskword ^ 0xffff)
    return maskwords


def _split_domain(domain, maxparts):
    """Return the parts of C{domain} as a list.

    Any parts beyond C{maxparts} are returned as a single part.

    _split_domain("foo.bar.baz", 2) => ["foo.bar", "baz"]

    """
    if maxparts == 1:
        return [domain]

    parts = domain.split(".")
    prefix = parts[:-maxparts + 1]
    suffix = parts[-maxparts + 1:]
    if prefix:
        return [".".join(prefix)] + suffix
    else:
        return parts


def _translate_domain(action, datadict, field, plain, maxparts, mapper):
    if field not in datadict:
        return
    try:
        plain = int(plain)
    except ValueError:
        raise ConfigError("Bad value for argument plain in action "
                          "{}: {}".format(action, repr(plain)))
    try:
        maxparts = int(maxparts)
    except ValueError:
        raise ConfigError("Bad value for argument maxparts in action "
                          "{}: {}".format(action, repr(maxparts)))

    parts = _split_domain(datadict[field], maxparts)

    for i in xrange(len(parts) - plain):
        parts[i] = mapper(parts[i:])

    datadict[field] = ".".join(parts)


def _random_domain_part():
    return "".join(random.choice("abcdefghijklmnopqrstuvwxyz"
                                 "ABCDEFGHIJKLMNOPQRSTUVWXYZ")
                   for _ in xrange(23))


def _pseudonymous_domain_part(parts, seed):
    if seed is None:
        raise ConfigError("Please set @domain_seed when using action "
                          "domain_pseudo")
    return hashlib.sha256(".".join(parts) + seed).hexdigest()
