# sorter.py
#
# Copyright (c) 2015 DFN-CERT Services GmbH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from collections import deque

from mixmod.errors import ConfigError
from mixmod import dataprocessor


class Sorter(dataprocessor.DataProcessor):
    """Sort mixmod data streams."""

    CONFIG = {
        "primary_sort_field": {
            "type": "string",
            "default": None
        },
        "secondary_sort_field": {
            "type": "string",
            "default": None
        },
        "tertiary_sort_field": {
            "type": "string",
            "default": None
        }
    }

    def __init__(self, **conf):
        self.datasets = list()
        self.primary_key = conf.pop("primary_sort_field", None)
        self.secondary_key = conf.pop("secondary_sort_field", None)
        self.tertiary_key = conf.pop("tertiary_sort_field", None)

        if self.primary_key is None:
            msg = 'Primary sort field not defined or config not found.'
            raise ConfigError(msg)

        self.results = deque()

    def put(self, mixmoddata):
        """
        This method adds a mixmod dataset to the Sorter.

        @type mixmoddata: dict MixModData
        @param mixmoddata: A MixModDataset dictionary that is part of list that
                           is to be sorted.
        """
        self.datasets.append(mixmoddata)

    def get(self):
        if self.results:
            return self.results.popleft()
        else:
            return None

    def close(self):
        if self.tertiary_key is not None:
            self.datasets = sorted(self.datasets,
                                   key=lambda dataset:
                                   dataset[self.tertiary_key])
        if self.secondary_key is not None:
            self.datasets = sorted(self.datasets,
                                   key=lambda dataset:
                                   dataset[self.secondary_key])
        self.datasets = sorted(self.datasets,
                               key=lambda dataset: dataset[self.primary_key])

        self.results = deque(self.datasets)
