# nfcap.py
#
# Copyright (c) 2010-2011, DFN-CERT Services GmbH
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


import logging
log = logging.getLogger('nfcap')
import struct
from mixmod.errors import DataError

__all__ = ["NFCap", "TCP_FLAG_ACK", "TCP_FLAG_RST", "TCP_FLAG_SYN",
           "TCP_FLAG_FIN", "TCP_FLAG_PSH", "TCP_FLAG_URG"]

# netinet/in.h
IPPROTO_ICMP = 1
IPPROTO_TCP = 6
IPPROTO_UDP = 17
IPPROTO_ICMPV6 = 58

TCP_FLAG_FIN = 1 << 0
TCP_FLAG_SYN = 1 << 1
TCP_FLAG_RST = 1 << 2
TCP_FLAG_PSH = 1 << 3
TCP_FLAG_ACK = 1 << 4
TCP_FLAG_URG = 1 << 5

NFCAP_MAGIC = 0xA50C
NFCAP_FLAG_COMPRESSED = 0x01
NFCAP_FLAG_EXTENDED_STATS = 0x02
NFCAP_FLAGS_MAX = NFCAP_FLAG_COMPRESSED | NFCAP_FLAG_EXTENDED_STATS

##################################################################
#
# nfcap file format description
#
# An nfcap file always starts with the same header. This is followed
# by a stats header (one of two types depending on a flag in the main
# header), which again is followed by a number of data blocks.
#
# +--------+--------------+--------------+-----+--------------+
# | Header | Stats Header | Data Block 1 | ... | Data Block n |
# +--------+--------------+--------------+-----+--------------+
#
# Each data block consists of a data block header followed by number
# of records. What kind of records depends on a flag in the data block
# header.
#
# Each record consists of a header followed by a number of additional
# data points. Which and what kind of data points depends on the flags
# in the record header.

# The global file header
nfcap_header = struct.Struct("<"
                             "H"  # u16 magic, NFCAP_MAGIC
                             "H"  # u16 version, should be 1 (v1.5 format)
                             "I"  # u32 flags, bitmask for see NFCAP_FLAG_*
                             "I"  # u32 numblocks, count of blocks in the file
                             "128s"  # ident, free-form string
                             )
# The simple stats header. There's also an extended stats header which
# is currently not supported.
nfcap_stats = struct.Struct("<"
                            "3Q"  # u64 total num flows / bytes / packets
                            "4Q"  # u64 num flows tcp / udp / icmp / other
                            "4Q"  # u64 num bytes tcp / udp / icmp / other
                            "4Q"  # u64 num packets tcp / udp / icmp / other
                            "I"  # u32 first seen, unix time stamp
                            "I"  # u32 last seen, unix time stamp
                            "H"  # u16 first seen milliseconds
                            "H"  # u16 last seen milliseconds
                            "I"  # u32 sequence failures, no clue what this is
                            )
# Each data block starts with this header.
nfcap_blockheader = struct.Struct("<"
                                  "I"   # u32 numrecords, number of records in
                                        # this block
                                  "I"   # u32 size, number of bytes in this
                                        # block excluding this header
                                  "H"   # u16 id, type of this block
                                        # 1 means v1.5 format
                                        # 2 means v1.6 format
                                  "H"   # u16 flags: 0=compat, 1=uncompressed, 2=compressed
                                  )
# Type 1 records in a data block
nfcap_record_v1 = struct.Struct("<"
                                "I"  # u32 flags, what extra info we have
                                "H"  # u16 size, size of this record in bytes
                                "H"  # u16 mark, ???
                                "H"  # u16 msec_first, start millisecs
                                "H"  # u16 msec_last, last millisecs
                                "I"  # u32 first, unix time stamp
                                "I"  # u32 last, unix time stamp
                                "BBBB"  # u8 dir, tcp_flags, prot, tos
                                "HH"  # u16 input / output, interface numbers
                                "HH"  # u16 srcport / dstport
                                "HH"  # u16 srcas / dstas, AS number
                                )

# In v1.5, each record header is followed by three fields: The src and
# dst IP address, the packet count, and the byte count. Those fields
# can be of varying sizes, though. Which sizes are used is defined by
# the bitmask of the flags field.
#
# bit 0 unset: IPv4 (2x32 bit) / set: IPv6 (2x128 bit)
# bit 1 unset: 32 bit packet count / set: 64 bit packet count
# bit 2 unset: 32 bit byte count / set: 64 bit byte count
nfcap_15_ipv4 = struct.Struct("<II")  # Two 32 bit IP numbers
nfcap_15_ipv6 = struct.Struct("<QQQQ")  # Two 128 bit IP numbers
nfcap_15_32bit = struct.Struct("<I")  # A 32 bit number
nfcap_15_64bit = struct.Struct("<Q")  # A 64 bit number

# Type 2 / v1.6
nfcap_record_v2_header = struct.Struct("<"
                                       "H"  # u16 record id
                                       "H"  # u16 record size
                                       )

nfcap_extension_map_record = struct.Struct("<"
                                    "H" # u16 map id
                                    "H" # u16 size
                                    "H" # u16 ref map id
                                    "H" # u16 ext size
                                    )

nfcap_record_v2_extension_map_header = struct.Struct("<"
                                                     "H"  # u16 map_id
                                                     "H"  # u16 extension_size
                                                     )
nfcap_record_v2_extension_map_entry = struct.Struct("<"
                                                    "H"  # u16 extension_id
                                                    )
nfcap_record_v2_common = struct.Struct("<"
                                       # "H" # u16 record_type
                                       # "H" # u16 size
                                       "B"  # u8 flags
                                       "B"  # u8 tag
                                       "H"  # u16 extension map
                                       "H"  # u16 msec_first
                                       "H"  # u16 msec_last
                                       "I"  # u32 first
                                       "I"  # u32 last
                                       "B"  # u8 fwd_status
                                       "B"  # u8 tcpflags
                                       "B"  # u8 proto
                                       "B"  # u8 src tos
                                       "H"  # u16 srcport
                                       "H"  # u16 dstport / ICMP
                                       )

# Record flag mapping
FLAG_MAP = [struct.Struct("<IIII"),    # 000 / ipv4, 32 bit packets, 32 bit bytes
            struct.Struct("<QQQQII"),  # 001 / ipv6, 32 bit packets, 32 bit bytes
            struct.Struct("<IIQI"),    # 010 / ipv4, 64 bit packets, 32 bit bytes
            struct.Struct("<QQQQQI"),  # 011 / ipv6, 64 bit packets, 32 bit bytes
            struct.Struct("<IIIQ"),    # 100 / ipv4, 32 bit packets, 64 bit bytes
            struct.Struct("<QQQQIQ"),  # 101 / ipv6, 32 bit packets, 64 bit bytes
            struct.Struct("<IIQQ"),    # 110 / ipv4, 64 bit packets, 64 bit bytes
            struct.Struct("<QQQQQQ"),  # 111 / ipv6, 64 bit packets, 64 bit bytes
            ]

ipv4 = struct.Struct("<I")
ipv6 = struct.Struct("<QQ")
ipv4x2 = struct.Struct("<II")
ipv6x2 = struct.Struct("<QQQQ")
u32 = struct.Struct("<I")
u64 = struct.Struct("<Q")

OP_DIRECT = 1
OP_IPV6 = 2


def labels(*args):
    return [(OP_DIRECT, index, name)
            for index, name in enumerate(args)]

EXTENSION_DESCRIPTIONS = {
    # Extension 4: Interfaces, 16 bit
    4: ("HH", labels("input", "output")),
    # Extension 5: Interfaces, 32 bit
    5: ("II", labels("input", "output")),
    # Extension 6: AS records, 16 bit
    6: ("HH", labels("srcas", "dstas")),
    # Extension 7: AS records, 32 bit
    7: ("II", labels("srcas", "dstas")),
    # Extension 8: Multiple fields
    8: ("BBBB", labels("tos", "dir", "srcmask", "dstmask")),
    #8: ("4x", []),
    # Extension 9: Next IP hop, v4
    9: ("I", labels("next_hop_ip")),
    #9: ("4x", []),
    # Extension 10: Next IP hop, v6
    10: ("QQ", [(OP_IPV6, 1, "next_hop_ip")]),
    #10: ("16x", []),
    # Extension 11: Next BGP hop, v4
    11: ("I", labels("next_bgp_ip")),
    #11: ("4x", []),
    # Extension 12: Next BGP hop, v6
    12: ("QQ", [(OP_IPV6, 1, "next_bgp_ip")]),
    #12: ("16x", []),
    # Extension 13: VLAN records
    13: ("HH", labels('srcvlan', 'dstvlan')),
    #13: ("4x", []),
    # Extension 14: Out packet counter, 32 bit
    14: ("I", labels('numoutpackets')),
    #14: ("4x", []),
    # Extension 15: Out packet counter, 64 bit
    15: ("Q", labels('numoutpackets')),
    #15: ("8x", []),
    # Extension 16: Out byte counter, 32 bit
    16: ("I", labels('numoutbytes')),
    #16: ("4x", []),
    # Extension 17: Out byte counter, 64 bit
    17: ("Q", labels('numoutbytes')),
    #17: ("8x", []),
    # Extension 18: Aggr flows, 32 bit
    18: ("I", labels('aggrflows')),
    #18: ("4x", []),
    # Extension 19: Aggr flows, 64 bit
    19: ("Q", labels('aggrflows')),
    #19: ("8x", []),
    # Extension 20: MAC addresses
    20: ("QQ", labels('in_src_mac', 'out_dst_mac')),
    #20: ("16x", []),
    # Extension 21: More MAC addresses
    21: ("QQ", labels('in_dst_mac', 'out_src_mac')),
    #21: ("16x", []),
    # Extension 22: MPLS labels
    22: ("IIIIIIIIII", labels('mpls_label_2', 'mpls_label_1',
                               'mpls_label_4', 'mpls_label_3',
                               'mpls_label_6', 'mpls_label_5',
                               'mpls_label_8', 'mpls_label_7',
                               'mpls_label_10', 'mpls_label_9')),
    #22: ("40x", []),
    # Extension 23: Router IPv4
    23: ("I", labels("router_ip")),
    #23: ("4x", []),
    # Extension 24: Router IPv6
    24: ("QQ", [(OP_IPV6, 1, "router_ip")]),
    #24: ("16x", []),
    # Extension 25: Router source ID
    25: ("HBB", labels('fill_e25', 'engine_type', 'engine_id')),
    #25: ("4x", []),
    }

# End of file format description
##################################################################


class NFCap(object):
    def __init__(self, rawdata):
        """
        Create an object representing an nfcap dump file.
        file should be a file object opened in binary mode.

        During initialization, this will read the nfcap header, but
        ignore the records for now. Call the iter_records method to
        iterate over the records.
        """
        #self.data = file.read()
        self.data = rawdata
        self.offset = nfcap_header.size + nfcap_stats.size
        self.read_header()
        self.ext_dict = dict()
        self.ext_list = list()

    def read_header(self):
        """
        Read the main file header and the stats block.
        This is done on object initialization
        """
        (magic, version, flags, numblocks,
         ident) = nfcap_header.unpack_from(self.data, 0)
        if magic != NFCAP_MAGIC:
            raise DataError("Bad MAGIC in file: %04x" % magic)
        if version != 1:
            raise DataError("Bad layout version %i, "
                             "only version 1 supported"
                             % (version,))
        if flags & NFCAP_FLAG_COMPRESSED:
            raise DataError("Compressed nfcap files not supported")
        if flags > NFCAP_FLAGS_MAX:
            raise DataError("Unknown flags: %04x" % (flags,))
        self.numblocks = numblocks
        self.ident = ident.replace("\x00", "")
        if flags & NFCAP_FLAG_EXTENDED_STATS:
            raise DataError("Extended stats not supported yet")
        else:
            self.read_stats()

    def read_stats(self):
        """
        Read the "old" stats header.
        """
        self.numflows = {}
        self.numbytes = {}
        self.numpackets = {}
        (self.numflows["total"], self.numbytes["total"],
         self.numpackets["total"],
         self.numflows["tcp"], self.numflows["udp"], self.numflows["icmp"],
         self.numflows["other"],
         self.numbytes["tcp"], self.numbytes["udp"], self.numbytes["icmp"],
         self.numbytes["other"],
         self.numpackets["tcp"], self.numpackets["udp"],
         self.numpackets["icmp"], self.numpackets["other"],
         self.first_seen, self.last_seen, self.msec_first, self.msec_last,
         self.sequence_failure) = nfcap_stats.unpack_from(self.data,
                                                          nfcap_header.size)

    def iter_records(self, wanted_proto=None, yield_none=False):
        """
        Iterate over all records in the file.
        """
        if wanted_proto is None:
            proto_mask = 0
            wanted_proto = 0
        else:
            proto_mask = -1
            # wanted_proto = wanted_proto

        data = self.data
        offset = self.offset
        extension_map = []
        for ign in xrange(self.numblocks):
            (numrecords, size, id, flags) = nfcap_blockheader.unpack_from(data, offset)
            offset += 12  # nfcap_blockheader.size
            if id == 2:
                for ign in xrange(numrecords):
                    value, offset = self.read_record_type_2(data,
                                                            offset,
                                                            extension_map,
                                                            proto_mask,
                                                            wanted_proto)
                    if value is False:  # Extension map or unknown record
                        continue
                    if value is not None or yield_none:
                        yield value
            elif id == 1:
                raise RuntimeError("Record type 1 not supported anymore")
            else:
                raise RuntimeError("Record type %s not supported" % id)


    def read_record_type_2(self, data, offset, extension_map,
                           proto_mask, wanted_proto):
        """
        Type 2 records are newer, v1.6 records.
        """
        (id, size) = nfcap_record_v2_header.unpack_from(data, offset)
        offset += 4  # nfcap_record_v2_header.size
        record_size = size - 4  # nfcap_record_v2_header.size
        final_offset = offset + record_size
        self.id_rec = id

        if id == 1:  # Common Record
            return (self.parse_record_type_2_common(record_size, data, offset,
                                                    extension_map,
                                                    proto_mask, wanted_proto),
                    final_offset)
        elif id == 2:  # Extension Map
            self.parse_extension_map(record_size, data, offset, extension_map)
        elif id == 0:  # Reserved
            log.warning("Record of type %i (reserved) ignored, %i bytes" %
                        (id, size))
        elif id == 3:  # Exporter
            log.warning("Record of type %i (exporter) ignored, %i bytes" %
                        (id, size))
        else:
            log.warning("Record of type %i (unknown) ignored, %i bytes" %
                        (id, size))
        return False, final_offset

    def parse_extension_map(self, record_size, data, offset, extension_map):
        (map_id, extension_size
         ) = nfcap_record_v2_extension_map_header.unpack_from(data, offset)
        offset += 4  # nfcap_record_v2_extension_map_header.size
        map_size = (record_size - 4) / 2
        map_struct = "<"
        map_labels = []
        map_labels_ipv6 = []
        map_index = 0

        for ext_id in struct.unpack_from("<" + ("H" * map_size),
                                         data,
                                         offset):

            if ext_id == 0:
                continue
            else:
                self.ext_list.append(ext_id)

        #print self.ext_list

        self.ext_dict[map_id] = self.ext_list
        self.ext_list = list()


    def parse_record_type_2_common(self, record_size, data, offset,
                                   extension_map, proto_mask, wanted_proto):
        r = Record()
        # TOT: 6s
        (flags, tag, extension_index, r.msec_first,
         r.msec_last, r.first, r.last, r.fwd_status, r.tcp_flags, proto, r.src_tos,
         r.src_port, r.dst_port) = nfcap_record_v2_common.unpack_from(data, offset)

        #if flags >= 1:
        #    print flags

        if proto & proto_mask != wanted_proto:
            return None
        r.prot = proto
        # TOT: 8.5s (+2.5s)

        offset += nfcap_record_v2_common.size # 24

        # Required extensions 1 (src/dst IP), 2 (in packets) and 3 (in bytes)
        # Optimized reading via a flag map
        st = FLAG_MAP[flags & 7]
        # TOT: 9.5s (+1s)
        values = st.unpack_from(data, offset)
        # TOT: 11s (+1.5s)
        offset += st.size
        if flags & 1:  # IPv6
            r.ipversion = 6
            (ip1a, ip1b, ip2a, ip2b,
             r.numpackets, r.numbytes) = values
            # << 64
            (r.srcip, r.dstip) = (ip1a * 18446744073709551616L + ip1b,
                                  ip2a * 18446744073709551616L + ip2b)
        else:  # IPv4
            r.ipversion = 4
            (r.srcip, r.dstip, r.numpackets, r.numbytes) = values
        # TOT: 12.5s (+1.5s)

        # first += msec_first / 1000.0
        # last += msec_last / 1000.0

        if proto == IPPROTO_ICMP:
            srcport = r.src_port
            # Fix odd CISCO behavior for ICMP port/type in src port
            if srcport != 0:
                r.icmptype = srcport % 256
                r.icmpcode = srcport / 256
            else:
                dstport = r.dst_port
                r.icmptype = dstport / 256
                r.icmpcode = dstport % 256
        else:
            r.icmptype, r.icmpcode = None, None

        for value in self.ext_dict[extension_index]:
            structure = struct.Struct("<" +
                EXTENSION_DESCRIPTIONS[value][0]
            )

            if value == 4 or value == 5:
                (r.input, r.output) = structure.unpack_from(data, offset)
                if value == 4:
                    offset += 4
                else:
                    offset += 8

            if value == 6 or value == 7:
                (r.src_as, r.dst_as) = structure.unpack_from(data, offset)

                if value == 6:
                    offset += 4
                else:
                    offset += 8

            if value == 8:
                (r.tos, r.dir, r.srcmask, r.dstmask) = structure.unpack_from(data, offset)
                offset += 4

            if value == 9:
                next_hop_ip = structure.unpack_from(data, offset)
                r.next_hop_ip = next_hop_ip[0]
                r.next_hop_ip_version = 4
                offset += 4

            if value == 10:
                (ip1a, ip1b) = structure.unpack_from(data, offset)
                r.next_hop_ip = (ip1a * 18446744073709551616L + ip1b)
                r.next_hop_ip_version = 6
                offset += 16

            if value == 11:
                next_bgp_ip = structure.unpack_from(data, offset)
                r.next_bgp_ip = next_bgp_ip[0]
                r.next_bgp_ip_version = 4
                offset += 4


            if value == 12:
                (ip2a, ip2b) = structure.unpack_from(data, offset)
                r.next_bgp_ip = (ip2a * 18446744073709551616L + ip2b)
                r.next_bgp_ip_version = 6
                offset += 16

            if value == 13:
                (r.srcvlan, r.dstvlan) = structure.unpack_from(data, offset)
                offset += 8

            if value == 14:
                r.numoutpackets = structure.unpack_from(data, offset)
                offset += 4

            if value == 15:
                r.numoutpackets = structure.unpack_from(data, offset)
                offset += 8

            if value == 16:
                r.numoutbytes = structure.unpack_from(data, offset)
                offset += 4

            if value == 17:
                r.numoutbytes = structure.unpack_from(data, offset)
                offset += 8

            if value == 18:
                r.aggrflows = structure.unpack_from(data, offset)
                offset += 4

            if value == 19:
                r.aggrflows = structure.unpack_from(data, offset)
                offset += 8

            if value == 20:
                (r.in_src_mac, r.out_dst_mac) = structure.unpack_from(data, offset)
                offset += 16

            if value == 21:
                (r.in_dst_mac, r.out_src_mac) = structure.unpack_from(data, offset)
                offset += 16

            if value == 22:
                (r.mpls_label_2, r.mpls_label_1, r.mpls_label_4,
                    r.mpls_label_3, r.mpls_label_6, r.mpls_label_5,
                    r.mpls_label_8, r.mpls_label_7, r.mpls_label_10,
                    r.mpls_label_9) = structure.unpack_from(data, offset)
                offset += 40

            if value == 23:
                router_ip = structure.unpack_from(data, offset)
                r.router_ip = router_ip[0]
                r.router_ip_version = 4
                offset += 4

            if value == 24:
                (ip3a, ip3b) = structure.unpack_from(data, offset)
                r.router_ip = (ip3a * 18446744073709551616L + ip3b)
                r.router_ip_version = 6
                offset += 16

            if value == 25:
                (r.fill_e25, r.engine_type, r.engine_id) = structure.unpack_from(data, offset)
                offset += 4

        return r


class Record(object):
    numflows = 1
    __slots__ = ('numpackets', 'numbytes',
                 'first', 'last', 'input', 'output',
                 'ipversion', 'prot', 'fwd_status',
                 'srcip', 'dstip', 'src_as', 'dst_as',
                 'tcp_flags', 'src_port', 'dst_port',
                 'icmptype', 'icmpcode',
                 'msec_first', 'msec_last', 'src_tos',
                 'tos', 'dir', 'srcmask', 'dstmask',
                 'next_hop_ip', 'next_bgp_ip', 'srcvlan', 'dstvlan',
                 'numoutpackets', 'numoutbytes', 'aggrflows',
                 'in_src_mac', 'out_dst_mac', 'in_dst_mac', 'out_src_mac',
                 'mpls_label_2', 'mpls_label_1', 'mpls_label_4', 'mpls_label_3',
                 'mpls_label_6', 'mpls_label_5', 'mpls_label_8', 'mpls_label_7',
                 'mpls_label_10', 'mpls_label_9', 'router_ip',
                 'fill_e25', 'engine_type', 'engine_id', 'next_hop_ip_version',
                 'next_bgp_ip_version', 'router_ip_version')

    def __repr__(self):
        return "<Record %s>" % ", ".join("%s=%s" % (name, getattr(self, name))
                                         for name in self.__slots__)

    def __getstate__(self):
        return tuple(getattr(self, name) for name in self.__slots__)

    def __setstate__(self, state):
        for name, value in zip(self.__slots__, state):
            setattr(self, name, value)
