# nfparse.py
#
# Copyright (c) 2011-2014, DFN-CERT Services GmbH
# All rights reserved.
# 
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
# 
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
# 
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.

# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.


import re

import nfcap
from netflowfilter import T, CIDR, ip2int, int2ip


##################################################################
# Main program

# usage:
# >>> data = nfparse.nfparse("...")
# >>> data.make_test()
# => T object for netflowfilter
# >>> data.make_json()
# => dictionary for json.dumps()

def nfparse(string):
    expr, tokens = expression(tokenize(string))
    if tokens.isempty():
        return expr
    else:
        raise ParsingError("Data after expression: %s" %
                           " ".join(tokens.tokens))


##################################################################
# Tokenizer

TOKEN_RX = re.compile(r'([a-zA-Z0-9:.,=<>/-]+|[\(\)])')


def tokenize(string):
    return Tokens(TOKEN_RX.findall(string))

##################################################################
# Grammar rules


def expression(tokens):
    # <expr> ::= ''
    if tokens.isempty():
        return TrueExpression(), tokens
    # <expr> ::= <or_expr>
    return or_expression(tokens)


def or_expression(tokens):
    # <or_expr> ::= <and_expr> or <or_expr>
    try:
        expr1, tokens2 = and_expression(tokens)
        tokens3 = expect("or", tokens2)
        expr2, tokens4 = or_expression(tokens3)
        return expr1 | expr2, tokens4
    except ParsingError:
        pass
    return and_expression(tokens)


def and_expression(tokens):
    # <and_expr> ::= <paren_expr> and <and_expr>
    try:
        expr1, tokens2 = paren_expression(tokens)
        tokens3 = expect("and", tokens2)
        expr2, tokens4 = and_expression(tokens3)
        return expr1 & expr2, tokens4
    except ParsingError:
        pass
    return paren_expression(tokens)


def paren_expression(tokens):
    try:
        # <paren_expr> ::= ( <expr> )
        tokens2 = expect("(", tokens)
        expr, tokens3 = expression(tokens2)
        tokens4 = expect(")", tokens3)
        return expr, tokens4
    except ParsingError:
        pass
    try:
        # <paren_expr> ::= not <paren_expr>
        tokens2 = expect("not", tokens)
        expr, tokens3 = paren_expression(tokens2)
        return ~expr, tokens3
    except ParsingError:
        pass
    try:
        # <paren_expr> ::= <statement>
        stmt, tokens2 = statement(tokens)
        return stmt, tokens2
    except ParsingError:
        pass
    raise ParsingError("Paren expression expected, found %r" %
                       tokens.next())


def statement(tokens):
    return one_of([statement_as, statement_bytes, statement_cidr,
                   statement_icmp_type, statement_icmp_code,
                   statement_ip, statement_ip_version, statement_interface,
                   statement_packets, statement_port, statement_protocol,
                   statement_tcp_flags, statement_country_code,
                   statement_start, statement_end], tokens)


def statement_as(tokens):
    # [src|dst] as [comp] <number>
    direction, tokens = token_direction(tokens)
    tokens = expect("as", tokens)
    comp, tokens = token_comparison(tokens)
    num, tokens = token_number(tokens)
    return FilterExpression("as", num,
                            direction=direction, comparison=comp), tokens


def statement_bytes(tokens):
    # bytes [<comp>] <unitnumber>
    tokens = expect("bytes", tokens)
    comp, tokens = token_comparison(tokens)
    num, tokens = token_unitnumber(tokens)
    return FilterExpression("numbytes", num, comparison=comp), tokens


def statement_cidr(tokens):
    # <cidr> ::= [src|dst] net <cidr>
    direction, tokens = token_direction(tokens)
    tokens = expect('net', tokens)
    cidr, tokens = token_cidr(tokens)
    return FilterExpression("ip", cidr,
                            comparison="in", direction=direction), tokens


def statement_icmp_type(tokens):
    # <icmp_type> ::= icmp-type <number>
    tokens = expect("icmp-type", tokens)
    num, tokens = token_number(tokens)
    return FilterExpression("icmptype", num), tokens


def statement_icmp_code(tokens):
    # <icmp_code> ::= icmp-code <number>
    tokens = expect("icmp-code", tokens)
    num, tokens = token_number(tokens)
    return FilterExpression("icmpcode", num), tokens


def statement_ip(tokens):
    # <ip> ::= [src|dst] (ip|host) [comp] <address>
    direction, tokens = token_direction(tokens)
    ignored, tokens = expect_any(["ip", "host"], tokens)
    comp, tokens = token_comparison(tokens)
    address, tokens = token_ip(tokens)
    return FilterExpression("ip", address,
                            direction=direction, comparison=comp), tokens


def statement_ip_version(tokens):
    # <ip_version> ::= inet | ipv4 | inet6 | ipv6
    ipversion, tokens = expect_any(["inet", "ipv4", "inet6", "ipv6"],
                                   tokens)
    if ipversion in ["inet", "ipv4"]:
        return FilterExpression("ipversion", 4), tokens
    elif ipversion in ["inet6", "ipv6"]:
        return FilterExpression("ipversion", 6), tokens
    else:
        raise RuntimeError("Can't happen (%r)" % (ipversion,))


def statement_interface(tokens):
    # <interface> ::= [in|out] if [comp] <number>
    direction, tokens = token_interfacedirection(tokens)
    tokens = expect("if", tokens)
    comp, tokens = token_comparison(tokens)
    num, tokens = token_number(tokens)
    return FilterExpression("if", num,
                            direction=direction, comparison=comp), tokens


def statement_packets(tokens):
    # <packets> ::= packets [comp] <unitnumber>
    tokens = expect("packets", tokens)
    comp, tokens = token_comparison(tokens)
    num, tokens = token_unitnumber(tokens)
    return FilterExpression("numpackets", num, comparison=comp), tokens


def statement_port(tokens):
    # <port> ::= [src|dst] port [comp] <number>
    direction, tokens = token_direction(tokens)
    tokens = expect("port", tokens)
    comp, tokens = token_comparison(tokens)
    num, tokens = token_number(tokens)
    return FilterExpression("port", num,
                            comparison=comp, direction=direction), tokens


IPPROTOCOLS = {"tcp": nfcap.IPPROTO_TCP,
               "udp": nfcap.IPPROTO_UDP,
               "icmp": nfcap.IPPROTO_ICMP,
               "icmpv6": nfcap.IPPROTO_ICMPV6}


def statement_protocol(tokens):
    # <protocol> ::= proto <number> | protocol <ipprotocol>
    tokens = expect("proto", tokens)
    try:
        proto_name, tokens = expect_any(IPPROTOCOLS.keys(), tokens)
        proto = IPPROTOCOLS[proto_name]
    except ParsingError:
        proto, tokens = token_number(tokens)
    return FilterExpression("prot", proto), tokens


def statement_tcp_flags(tokens):
    # <tcp_flags> ::= flags <flags>
    tokens = expect("flags", tokens)
    flags, tokens = token_flags(tokens)
    return FilterExpression("tcp_flags", flags, comparison="bitmask"), tokens


def statement_country_code(tokens):
    # <country_code> ::= [src|dst] country <country_code>
    direction, tokens = token_direction(tokens)
    tokens = expect("country", tokens)
    country_code, tokens = token_country_code(tokens)
    return FilterExpression("country_code", country_code,
                            direction=direction), tokens


def statement_start(tokens):
    # <start> ::= start [comp] <number>
    tokens = expect("start", tokens)
    comp, tokens = token_comparison(tokens)
    start, tokens = token_number(tokens)
    return FilterExpression("start", start,
                            comparison=comp), tokens


def statement_end(tokens):
    # <end> ::= end [comp] <number>
    tokens = expect("end", tokens)
    comp, tokens = token_comparison(tokens)
    end, tokens = token_number(tokens)
    return FilterExpression("end", end,
                            comparison=comp), tokens

##################################################################
# Terminals / tokens


def token_number(tokens):
    # <number> ::= [0-9]+
    next, tokens2 = tokens.consume()
    if next.isdigit():
        return int(next), tokens2
    else:
        raise ParsingError("Number expected, %r found" % next)


def token_unitnumber(tokens):
    # <unitnumber> ::= <number>[kmg]?
    next, tokens2 = tokens.consume()
    if next[-1].lower() in ['k', 'm', 'g']:
        factor = {'k': 1000, 'm': 1000 ** 2, 'g': 1000 ** 3}[next[-1].lower()]
        next = next[:-1]
    else:
        factor = 1
    if next.isdigit():
        return factor * int(next), tokens2
    else:
        raise ParsingError("Number with unit expected, %r found" % next)


def reverse_units(num):
    for factor, unit in [(1000 ** 3, 'g'), (1000 ** 2, 'm'), (1000 ** 1, 'k')]:
        if ((num / factor) * factor) == num:
            return "%s%s" % (num / factor, unit)
    return str(num)


def token_comparison(tokens):
    # <comp> ::= = | == | > | < | eq | lt | gt
    # default eq
    next, tokens2 = tokens.consume()
    if next in ['==', '=', 'eq']:
        return 'eq', tokens2
    elif next in ['>', 'gt']:
        return 'gt', tokens2
    elif next in ['<', 'lt']:
        return 'lt', tokens2
    elif next in ['<=', 'lte']:
        return 'lte', tokens2
    elif next in ['>=', 'gte']:
        return 'gte', tokens2
    elif next in ['!=', 'ne', 'neq']:
        return 'ne', tokens2
    else:
        return 'eq', tokens


def token_direction(tokens):
    # src | dst, default = "any"
    if tokens.next() in ["src", "dst"]:
        return tokens.consume()
    else:
        return "any", tokens


def token_interfacedirection(tokens):
    # in | out, default = "any"
    if tokens.next() in ["in", "out"]:
        return tokens.consume()
    else:
        return "any", tokens


def token_cidr(tokens):
    next, tokens = tokens.consume()
    try:
        return CIDR(next), tokens
    except:
        raise ParsingError("Expected a CIDR address, found %r" % next)


def token_ip(tokens):
    next, tokens = tokens.consume()
    try:
        return ip2int(next), tokens
    except:
        raise ParsingError("Expected an IP address, found %r" % next)


def token_flags(tokens):
    next, tokens = tokens.consume()
    next = next.lower()
    if next.strip("asfr") != '':
        raise ParsingError("Expected TCP flags argument, found %r" % next)
    result = 0
    if "a" in next:
        result |= nfcap.TCP_FLAG_ACK
    if "s" in next:
        result |= nfcap.TCP_FLAG_SYN
    if "f" in next:
        result |= nfcap.TCP_FLAG_FIN
    if "r" in next:
        result |= nfcap.TCP_FLAG_RST
    if "p" in next:
        result |= nfcap.TCP_FLAG_PSH
    if "u" in next:
        result |= nfcap.TCP_FLAG_URG
    return result, tokens


def token_country_code(tokens):
    next, tokens = tokens.consume()
    next = next.upper()
    if next.isalpha() and len(next) == 2:
        return next, tokens
    else:
        raise ParsingError("Invalid country code %s found" % next)

##################################################################
# Utilities


class Tokens(object):
    def __init__(self, tokens):
        self.tokens = tokens

    def isempty(self):
        return len(self.tokens) == 0

    def next(self):
        if self.isempty():
            raise ParsingError("Attempted to read past the input stream")
        return self.tokens[0]

    def consume(self):
        if self.isempty():
            raise ParsingError("Attempted to read past the input stream")
        return self.tokens[0], Tokens(self.tokens[1:])


def expect(literal, tokens):
    next, tokens2 = tokens.consume()
    if next == literal:
        return tokens2
    raise ParsingError("Expected %r, found %r" % (literal, next))


def expect_any(literals, tokens):
    next, tokens2 = tokens.consume()
    for literal in literals:
        if next == literal:
            return literal, tokens2
    raise ParsingError("Expected one of %s, found %r" % (", ".join(literals),
                                                         next))


def one_of(expansions, tokens):
    for expansion in expansions:
        try:
            return expansion(tokens)
        except ParsingError:
            pass
    raise ParsingError("Expected an expansion, found %r" % tokens.next())


class ParsingError(Exception):
    pass


##################################################################
# Output

class Expression(object):
    def __and__(self, arg):
        return AndExpression(self, arg)

    def __or__(self, arg):
        return OrExpression(self, arg)

    def __invert__(self):
        return NotExpression(self)

    def proto(self):
        """
        Return a protocol we filter for, or None if none.

        By default (And, Not) we do not have a specific such option.
        """
        return None

    def simplify(self):
        """Return a simplified form of this expression."""
        return self

    def disjunctive_normal_form(self):
        """Return an equivalent form in disjunctive normal form."""
        return self


class TrueExpression(Expression):
    def make_test(self):
        return T()

    def make_json(self):
        return {}

    def make_nfparse(self):
        # Shouldn't happen within an expression.
        return ""


class BinaryExpression(Expression):
    def __init__(self, expr1, expr2):
        self.expr1 = expr1
        self.expr2 = expr2

    def __eq__(self, other):
        if not isinstance(other, type(self)):
            return False
        return self.expr1 == other.expr1 and self.expr2 == other.expr2

    def _operands(self):
        """Return a list of operands at the same level as this."""
        agenda = [self]
        operands = []
        while len(agenda) > 0:
            (this, agenda) = (agenda[0], agenda[1:])
            if isinstance(this, type(self)):
                agenda.insert(0, this.expr1)
                agenda.append(this.expr2)
            else:
                operands.append(this)
        return operands

    def simplify(self):
        # Remove duplicates while retaining order
        operands = []
        for operand in self._operands():
            operand = operand.simplify()
            if operand not in operands:
                operands.append(operand)
        # A x (A y b) == A for x, y \in AND, OR

        def is_superfluous(expr1, expr2):
            if not isinstance(expr2, BinaryExpression):
                return False
            elif isinstance(expr2, type(self)):
                return False
            elif expr1 in expr2._operands():
                return True
            else:
                return False
        superfluous = []
        for expr1 in operands:
            for expr2 in operands:
                if expr1 != expr2 and is_superfluous(expr1, expr2):
                    superfluous.append(expr2)
        operands = [operand for operand in operands
                    if operand not in superfluous]

        expr = operands.pop()
        while operands:
            expr = type(self)(operands.pop(), expr)
        return expr


class AndExpression(BinaryExpression):
    def __str__(self):
        return "(%s and %s)" % (self.expr1, self.expr2)

    def make_test(self):
        return self.expr1.make_test() & self.expr2.make_test()

    def make_json(self):
        conjuncts = self._operands()
        return {'type': 'group',
                'connective': 'and',
                'invert': False,
                'contents': [expression.make_json()
                             for expression in conjuncts]}

    def make_nfparse(self):
        conjuncts = []
        for operand in self._operands():
            if isinstance(operand, OrExpression):
                conjuncts.append("({})".format(operand.make_nfparse()))
            else:
                conjuncts.append(operand.make_nfparse())
        return " and ".join(conjuncts)

    def proto(self):
        """
        Return a protocol we filter for.

        For an AndExpression, this is either of the two.
        """
        return self.expr1.proto() or self.expr2.proto() or None

    def disjunctive_normal_form(self):
        expr = self.simplify()
        if not isinstance(expr, AndExpression):
            return expr.disjunctive_normal_form()
        # Distributive law
        # We use the distributive law iff one of our constituents is
        # a disjunct.
        expr1 = expr.expr1.disjunctive_normal_form()
        expr2 = expr.expr2.disjunctive_normal_form()
        if isinstance(expr1, OrExpression):
            newexpr = OrExpression(AndExpression(expr1.expr1, expr2),
                                   AndExpression(expr1.expr2, expr2))
            return newexpr.disjunctive_normal_form()
        elif isinstance(expr2, OrExpression):
            newexpr = OrExpression(AndExpression(expr1, expr2.expr1),
                                   AndExpression(expr1, expr2.expr2))
            return newexpr.disjunctive_normal_form()
        else:
            return expr


class OrExpression(BinaryExpression):
    def __str__(self):
        return "(%s or %s)" % (self.expr1, self.expr2)

    def make_test(self):
        return self.expr1.make_test() | self.expr2.make_test()

    def make_json(self):
        disjuncts = self._operands()
        return {'type': 'group',
                'connective': 'or',
                'invert': False,
                'contents': [expression.make_json()
                             for expression in disjuncts]}

    def make_nfparse(self):
        disjuncts = []
        for operand in self._operands():
            if isinstance(operand, AndExpression):
                disjuncts.append("({})".format(operand.make_nfparse()))
            else:
                disjuncts.append(operand.make_nfparse())
        return " or ".join(disjuncts)

    def disjunctive_normal_form(self):
        expr = self.simplify()
        if not isinstance(expr, OrExpression):
            return expr.disjunctive_normal_form()
        return OrExpression(expr.expr1.disjunctive_normal_form(),
                            expr.expr2.disjunctive_normal_form())


class NotExpression(Expression):
    def __init__(self, expr):
        self.expr = expr

    def __str__(self):
        return "not %s" % (self.expr,)

    def __eq__(self, other):
        if not isinstance(other, NotExpression):
            return False
        return self.expr == other.expr

    def simplify(self):
        """Return a simplified form of this expression."""
        expr = self.expr.simplify()
        # Duplicate not eliminates itself
        if isinstance(expr, NotExpression):
            return expr.expr
        else:
            return NotExpression(expr)

    def make_test(self):
        return ~self.expr.make_test()

    def make_json(self):
        expr = self.expr.make_json()
        expr['invert'] = True
        return expr

    def make_nfparse(self):
        if isinstance(self.expr, BinaryExpression):
            return "not ({})".format(self.expr.make_nfparse())
        else:
            return "not " + self.expr.make_nfparse()

    def disjunctive_normal_form(self):
        expr = self.expr.disjunctive_normal_form()
        # De Morgan
        if isinstance(expr, AndExpression):
            expr = self.simplify().expr
            return OrExpression(
                NotExpression(expr.expr1).disjunctive_normal_form(),
                NotExpression(expr.expr2).disjunctive_normal_form()
            ).disjunctive_normal_form()
        elif isinstance(self.expr, OrExpression):
            expr = self.simplify().expr
            return AndExpression(
                NotExpression(expr.expr1).disjunctive_normal_form(),
                NotExpression(expr.expr2).disjunctive_normal_form()
            ).disjunctive_normal_form()
        else:
            return NotExpression(expr).simplify()


class FilterExpression(Expression):
    FIELDMAP = {('src', 'as'): ['srcas'],
                ('dst', 'as'): ['dstas'],
                ('any', 'as'): ['srcas', 'dstas'],
                ('src', 'country_code'): ['srccountry'],
                ('dst', 'country_code'): ['dstcountry'],
                ('any', 'country_code'): ['srccountry', 'dstcountry'],
                ('start', 'time'): ['start'],
                ('end', 'time'): ['end'],
                ('src', 'ip'): ['srcip'],
                ('dst', 'ip'): ['dstip'],
                ('any', 'ip'): ['srcip', 'dstip'],
                ('src', 'port'): ['srcport'],
                ('dst', 'port'): ['dstport'],
                ('any', 'port'): ['srcport', 'dstport'],
                ('in', 'if'): ['input'],
                ('out', 'if'): ['output'],
                ('any', 'if'): ['input', 'output'],
                }

    def __init__(self, field, value, comparison="eq", direction=None):
        self.direction = direction
        self.field = field
        self.comparison = comparison
        self.value = value

    def __str__(self):
        s = ""
        if self.direction is not None and self.direction != 'any':
            s += "%s " % self.direction
        s += "%s " % self.field
        if self.comparison != 'eq':
            s += "%s " % self.comparison
        s += "%s" % self.value
        return s

    def __eq__(self, other):
        if not isinstance(other, FilterExpression):
            return False
        return (self.direction == other.direction and
                self.field == other.field and
                self.comparison == other.comparison and
                self.value == other.value)

    def make_test(self):
        if self.direction is None:
            fields = [self.field]
        else:
            fields = self.FIELDMAP[(self.direction, self.field)]
        test = self.make_single_test(fields[0])
        for field in fields[1:]:
            test |= self.make_single_test(field)
        return test

    def make_single_test(self, field):
        if self.comparison == 'eq':
            return T(**{field: self.value})
        else:
            return T(**{"%s__%s" % (field, self.comparison): self.value})

    def make_json(self):
        field = self.field
        value = self.value
        if field == 'ip':
            # value is either a CIDR object or an integer
            if isinstance(value, CIDR):
                field = 'net'
                value = value.name
            elif isinstance(value, (int, long)):
                value = int2ip(value)
            else:
                value = value
        elif field == 'tcp_flags':
            field = 'flags'
            value = ""
            if self.value & nfcap.TCP_FLAG_ACK:
                value += "A"
            if self.value & nfcap.TCP_FLAG_SYN:
                value += "S"
            if self.value & nfcap.TCP_FLAG_FIN:
                value += "F"
            if self.value & nfcap.TCP_FLAG_RST:
                value += "R"
            if value == "":  # Shouldn't happen
                value = self.value
        elif field == 'numbytes':
            field = 'bytes'
            value = reverse_units(value)
        elif field == 'numpackets':
            field = 'packets'
            value = reverse_units(value)
        elif field == 'prot':
            field = 'proto'
        elif field == 'icmptype':
            field = 'icmp-type'
        elif field == 'icmpcode':
            field = 'icmp-code'

        return {'type': 'statement',
                'invert': False,
                'direction': self.direction,
                'field': field,
                'comparison': self.comparison,
                'value': str(value)}

    def make_nfparse(self):
        expr = []
        field, value = self._unparse_field_value()
        if self.direction is not None and self.direction != 'any':
            expr.append(self.direction)
        expr.append(field)
        if self.comparison not in (None, 'eq', 'in', 'bitmask'):
            expr.append({'gt': '>',
                         'lt': '<',
                         'gte': '>=',
                         'lte': '<=',
                         'ne': '!='}[self.comparison])
        if value is not None:
            expr.append(str(value))
        return " ".join(expr)

    def _unparse_field_value(self):
        field = self.field
        value = self.value
        if field == 'ip':
            # value is either a CIDR object or an integer
            if isinstance(value, CIDR):
                return 'net', value.name
            elif isinstance(value, (int, long)):
                return field, int2ip(value)
            else:
                return field, value
        elif field == 'tcp_flags':
            field = 'flags'
            value = ""
            if self.value & nfcap.TCP_FLAG_ACK:
                value += "A"
            if self.value & nfcap.TCP_FLAG_SYN:
                value += "S"
            if self.value & nfcap.TCP_FLAG_FIN:
                value += "F"
            if self.value & nfcap.TCP_FLAG_RST:
                value += "R"
            if value == "":  # Shouldn't happen
                value = self.value
            return field, value
        elif field == 'numbytes':
            return 'bytes', reverse_units(value)
        elif field == 'numpackets':
            return 'packets', reverse_units(value)
        elif field == 'prot':
            for name, num in IPPROTOCOLS.items():
                if num == value:
                    value = name
                    break
            return 'proto', value
        elif field == 'icmptype':
            return 'icmp-type', value
        elif field == 'icmpcode':
            return 'icmp-code', value
        elif field == 'ipversion':
            return 'ipv' + str(value), None
        else:
            return field, value

    def proto(self):
        """
        Return a protocol we filter for.

        Only if we filter for a field.
        """
        if self.field == 'prot':
            return self.value
        else:
            return None


# Expansion rules:
#
# <expr> ::= <or_expr>
# <or_expr> ::= <and_expr> or <or_expr>
# <and_expr> ::= <paren_expr> and <and_expr>
# <paren_expr> ::= ( <expr> )
# <paren_expr> ::= not <paren_expr>
# <paren_expr> ::= <statement>
#
# <statement> ::= <as> | <bytes> | <cidr> | <icmp_type> | <icmp_code> |
#                 <ip> | <ip_version> | <interface> | <packets> |
#                 <port> | <protocol> | <tcp_flags>
#
# <as> ::= [src|dst] as [comp] <number>
# <bytes> ::= bytes [<comp>] <unitnumber>
# <cidr> ::= [src|dst] net <cidr>
# <icmp_type> ::= icmp-type <number>
# <icmp_code> ::= icmp-code <number>
# <ip> ::= [src|dst] (ip|host) [comp] <address>
# <ip_version> ::= inet | ipv4 | inet6 | ipv6
# <interface> ::= [in|out] if [comp] <number>
# <packets> ::= packets [comp] <unitnumber>
# <port> ::= [src|dst] port [comp] <number>
# <protocol> ::= proto <number> | protocol <ipprotocol>
# <tcp_flags> ::= flags <flags>
#
# Terminals:
#
# <cidr> ::= <address>/<number>
# <address> ::= [0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\.[0-9]{1,3}\
# <comp> ::= = | == | > | < | eq | lt | gt
# <number> ::= [0-9]+
# <unitnumber> ::= <number>[kmg]?
# <ipprotocol> ::= tcp | udp | icmp
# <flags> ::= [ASFR]+
