import unittest

import netaddr

from mixmod.modules import pseudonymizer


class PseudonymizeTestCase(unittest.TestCase):
    def create(self, **conf):
        return pseudonymizer.Pseudonymizer(**conf)


class IPTestCase(PseudonymizeTestCase):
    def translate(self, field, algorithm, value):
        data = {field: value}
        p = self.create(
            **{field: algorithm,
               "@aes_key_set": "this_is_a_32_byte_string123456789"}
        )
        p.put(data)
        return p.get()


class IPPseudonomyousTestCase(IPTestCase):
    algorithm = "ip_aes_hash"

    def test_should_translate_same_ip4s_to_same_addresses(self):
        translated1 = self.translate(
            field="dst_ip_v4",
            algorithm=self.algorithm,
            value=netaddr.IPAddress("10.11.12.13")
        )
        translated2 = self.translate(
            field="dst_ip_v4",
            algorithm=self.algorithm,
            value=netaddr.IPAddress("10.11.12.13")
        )

        self.assertEqual(translated1["dst_ip_v4"],
                         translated2["dst_ip_v4"])

    def test_should_translate_different_ip4s_to_different_addresses(self):
        translated1 = self.translate(
            field="dst_ip_v4",
            algorithm=self.algorithm,
            value=netaddr.IPAddress("10.11.12.13")
        )
        translated2 = self.translate(
            field="dst_ip_v4",
            algorithm=self.algorithm,
            value=netaddr.IPAddress("10.11.12.14")
        )

        self.assertNotEqual(translated1["dst_ip_v4"],
                            translated2["dst_ip_v4"])

    def test_should_mark_src_data_as_pseudonymous(self):
        translated = self.translate(
            field="src_ip_v4",
            algorithm=self.algorithm,
            value=netaddr.IPAddress("10.11.12.13")
        )

        self.assertEqual(translated["src_mode"], 2)

    def test_should_mark_dst_data_as_pseudonymous(self):
        translated = self.translate(
            field="dst_ip_v4",
            algorithm=self.algorithm,
            value=netaddr.IPAddress("10.11.12.13")
        )

        self.assertEqual(translated["dst_mode"], 2)


class TestIPAESHash(IPPseudonomyousTestCase):
    algorithm = 'ip_map'


class TestDomainAnon(PseudonymizeTestCase):
    def translate(self, field, algorithm, value):
        data = {field: value}
        p = self.create(**{field: algorithm})
        p.put(data)
        return p.get()

    def test_should_anonymize_domain(self):
        datadict = self.translate("foo", "domain_anon", "www.frobnicate.test")
        actual = datadict["foo"].split(".")

        self.assertEqual(len(actual), 3)
        self.assertNotEqual(actual[0], "www")
        self.assertNotEqual(actual[1], "frobnicate")
        self.assertNotEqual(actual[2], "test")

    def test_should_anonymize_domain_keeping_plain_parts(self):
        datadict = self.translate("foo", "domain_anon plain=1",
                                  "www.frobnicate.test")
        actual = datadict["foo"].split(".")

        self.assertEqual(len(actual), 3)
        self.assertNotEqual(actual[0], "www")
        self.assertNotEqual(actual[1], "frobnicate")
        self.assertEqual(actual[2], "test")

    def test_should_anonymize_domain_merging_parts(self):
        datadict = self.translate("foo", "domain_anon plain=1 maxparts=2",
                                  "www.frobnicate.test")
        actual = datadict["foo"].split(".")

        self.assertEqual(len(actual), 2)
        self.assertNotEqual(actual[0], "www.frobnicate")
        self.assertEqual(actual[1], "test")


class TestDomainPseudo(PseudonymizeTestCase):
    def translate(self, field, algorithm, value):
        data = {field: value}
        p = self.create(**{"@domain_seed": "12345",
                           field: algorithm})
        p.put(data)
        return p.get()

    def test_should_pseudonymize_domain(self):
        datadict1 = self.translate("foo", "domain_pseudo",
                                   "www.frobnicate.test")
        actual1 = datadict1["foo"].split(".")
        datadict2 = self.translate("foo", "domain_pseudo",
                                   "www.frobnicate.test")
        actual2 = datadict2["foo"].split(".")

        self.assertEqual(len(actual1), 3)
        self.assertNotEqual(actual1[0], "www")
        self.assertNotEqual(actual1[1], "frobnicate")
        self.assertNotEqual(actual1[2], "test")
        self.assertEqual(actual1, actual2)

    def test_should_pseudonymize_domain_keeping_plain_parts(self):
        datadict1 = self.translate("foo", "domain_pseudo plain=1",
                                   "www.frobnicate.test")
        actual1 = datadict1["foo"].split(".")
        datadict2 = self.translate("foo", "domain_pseudo plain=1",
                                   "www.frobnicate.test")
        actual2 = datadict2["foo"].split(".")

        self.assertEqual(len(actual1), 3)
        self.assertNotEqual(actual1[0], "www")
        self.assertNotEqual(actual1[1], "frobnicate")
        self.assertEqual(actual1[2], "test")
        self.assertEqual(actual1, actual2)

    def test_should_pseudonymize_domain_merging_parts(self):
        datadict1 = self.translate("foo", "domain_pseudo plain=1 maxparts=2",
                                   "www.frobnicate.test")
        actual1 = datadict1["foo"].split(".")
        datadict2 = self.translate("foo", "domain_pseudo plain=1 maxparts=2",
                                   "www.frobnicate.test")
        actual2 = datadict2["foo"].split(".")

        self.assertEqual(actual1, actual2)
        self.assertEqual(len(actual1), 2)
        self.assertNotEqual(actual1[0], "www.frobnicate")
        self.assertEqual(actual1[1], "test")

    def test_should_not_allow_pseudonymize_parts(self):
        datadict1 = self.translate("foo", "domain_pseudo plain=1",
                                   "www.foo.test")
        actual1 = datadict1["foo"].split(".")
        datadict2 = self.translate("foo", "domain_pseudo plain=1",
                                   "www.bar.test")
        actual2 = datadict2["foo"].split(".")

        self.assertNotEqual(actual1[0], actual2[0])
        self.assertNotEqual(actual1[1], actual2[1])
        self.assertEqual(actual1[2], actual2[2])


class TestSplitDomain(unittest.TestCase):
    def test_should_split_all_for_zero(self):
        self.assertEqual(["foo", "bar", "baz"],
                         pseudonymizer._split_domain("foo.bar.baz", 0))

    def test_should_not_split_for_one(self):
        self.assertEqual(["foo.bar.baz"],
                         pseudonymizer._split_domain("foo.bar.baz", 1))

    def test_should_split_for_two(self):
        self.assertEqual(["foo.bar", "baz"],
                         pseudonymizer._split_domain("foo.bar.baz", 2))

    def test_should_split_for_larger(self):
        self.assertEqual(["foo", "bar", "baz"],
                         pseudonymizer._split_domain("foo.bar.baz", 15))
