/*  aapimodule.c
 
 Copyright (c) 2014, DFN-CERT Services GmbH
 All rights reserved.
 
 Redistribution and use in source and binary forms, with or without
 modification, are permitted provided that the following conditions are met:
 
 1. Redistributions of source code must retain the above copyright notice,
 this list of conditions and the following disclaimer.
 
 2. Redistributions in binary form must reproduce the above copyright notice,
 this list of conditions and the following disclaimer in the documentation
 and/or other materials provided with the distribution.
 THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
 AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
 LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 POSSIBILITY OF SUCH DAMAGE.
*/


#include <Python.h>

#include <stdlib.h>
#include <string.h>
#include <time.h>
#include <arpa/inet.h>
#include <netinet/in.h>

#include "../lib/anonymization_functions.h"
#include "../lib/aes.h"
#include "../lib/prefix_preserving_map.h"

#include "aapimodule.h"
#include "anon_prefix_preserving.h"

static char *statebuf;

/* -===================================================================-
 * mapping tables 
 */
static nodehdr_t addr_propagate = { NH_FL_RANDOM_PROPAGATE, 0xffffffff, 0x01000000 };
 
static mapNode *srcIpMappingTable[MAPPING_ENTRIES];
static int src_ip_count = 16777217;
static mapNode *dstIpMappingTable[MAPPING_ENTRIES];
static int dst_ip_count = 1107296257;

static mapNode *ipMappingTable[MAPPING_ENTRIES];
static int ip_count = 16777217;

static mapNode *srcIp6MappingTable[MAPPING_ENTRIES];
static int src_ip6_count = 16777217;
static mapNode *dstIp6MappingTable[MAPPING_ENTRIES];
static int dst_ip6_count = 1107296257;


static mapNode *ip6MappingTable[MAPPING_ENTRIES];
static int ip6_count = 16777217;

unsigned char AES_keys[32] = { 
  0x01, 0x23, 0x45, 0x67, 0x89, 0xAB, 0xCD, 0xEF, 0x23, 0x45,
  0x67, 0x89, 0xAB, 0xCD, 0xEF, 0x01, 0x45, 0x67, 0x89, 0xAB, 
  0xCD, 0xEF, 0x01, 0x23, 0x01, 0x23, 0x45, 0x67, 0x23, 0x45, 
  0x67, 0x89
};

/* -===================================================================-
 *  Python bindings
 */
 
static PyMethodDef AAPIMethods[] = {
  {"aes_key_set", aes_key_set, METH_VARARGS,
    "Set a new AES key for hashing functions"},
  
  {"ip_aes_hash", ip_aes_hash, METH_VARARGS,
    "AES hash an IP address"},
  {"ip_bd_map", ip_bd_map, METH_VARARGS,
    "Bidirectional mapping of src and dst IP addresses"},
  {"ip_map", ip_map, METH_VARARGS, 
    "Alter an IP according to a dynamic mapping table"},
  {"ip_pref_pres_anon", ip_pref_pres_anon, METH_VARARGS,
    "Anonymize an IP preserving its prefix"},
  {"ip_pref_pres_map", ip_pref_pres_map, METH_VARARGS,
    "Alter an IP according to a mapping table, preserving its prefix"},
  {"ip_random", ip_random, METH_VARARGS, 
    "Substitute IP bits with random values"},
  {"ip_truncate", ip_truncate, METH_VARARGS,
    "Set the last n bits of an IP address to 0"},
  
  {"ip6_aes_hash", ip6_aes_hash, METH_VARARGS,
    "AES hash an IPv6 address"},
  {"ip6_bd_map", ip6_bd_map, METH_VARARGS,
    "Bidirectional mapping of src and dst IPv6 addresses"},
  {"ip6_map", ip6_map, METH_VARARGS,
    "Alter an IPv6 address according to a dynamic mapping table"},
  
  {NULL, NULL, 0, NULL}
};

PyMODINIT_FUNC initaapi(void) {
  (void) Py_InitModule ("aapi", AAPIMethods);
  
  if (!(statebuf = malloc(128))) {
    PyErr_SetString(PyExc_MemoryError, "Out of memory");
    return;
  }
    
  /* PRNG */
  initstate(time(NULL), statebuf, 128);
  srandom(time(NULL));

  /* mapping tables */
  memset(ipMappingTable, 0, MAPPING_ENTRIES * sizeof(mapNode *));
  memset(srcIpMappingTable, 0, MAPPING_ENTRIES * sizeof(mapNode *));
  memset(dstIpMappingTable, 0, MAPPING_ENTRIES * sizeof(mapNode *));
  
  lookup_init(&addr_propagate);
  
}

/* -===================================================================-
 *  Actual implementation follows
 */

static PyObject *aes_key_set(PyObject *self, PyObject *args) {

  /* "You must not provide storage for the string itself; a pointer to an
   * existing string is stored into the character pointer variable whose
   * address you pass" - https://docs.python.org/2/c-api/arg.html
   */
  char *input_string = NULL;
  
  if (! PyArg_ParseTuple(args, "s", &input_string)) {
    return NULL;
  }
  if (strlen(input_string) < 32) {
    PyErr_SetString(PyExc_ValueError, "need 32 byte of new key material");
    return NULL;
  }
  
  strncpy((char *)AES_keys, input_string, 32);
  
  return Py_BuildValue("s#", AES_keys, 32);
}

#if 0
static PyObject *ip_aes_hash(PyObject *self, PyObject *args) {
  uint32_t ip;
  
  if (! PyArg_ParseTuple(args, "k", &ip)) {
    return NULL;
  }
  ip = htonl(ip);
  
  aes_hash((unsigned char *)(&ip), 4, 
    (unsigned char *)AES_keys, 8 * sizeof(AES_keys),
    0, NULL);
  
  return Py_BuildValue("k", ntohl(ip));
}
#else
static PyObject *ip_aes_hash(PyObject *self, PyObject *args) {
  uint32_t ip;
  uint32_t *result;
  unsigned char *field;
  static aes_context ctx;
  unsigned char inbuf[16] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
  unsigned char outbuf[16] = { 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
  int i;
  
  if (! PyArg_ParseTuple(args, "k", &ip)) {
    return NULL;
  }
  ip = htonl(ip);
  
  field = (unsigned char *)(&ip);
  aes_schedule_key(&ctx, (unsigned char *)AES_keys, 8 * sizeof(AES_keys));

  for (i = 0; i < 4; i ++)
    inbuf[i] = field[i];

  aes_encrypt(&ctx, inbuf, outbuf);

  for (i = 0; i < 4; i ++)
    field[i] = outbuf[i];
  
  result = ((uint32_t *)(field));
  
  /* return Py_BuildValue("kkkk", ntohl(result[0]),
       ntohl(result[1]), ntohl(result[2]), ntohl(result[3]));
   */
 return Py_BuildValue("k", ntohl(result[0]));
}
#endif

static PyObject *ip_bd_map(PyObject *self, PyObject *args) {
  uint32_t srcip, dstip;
  
  if (! PyArg_ParseTuple(args, "(kk)", &srcip, &dstip)) {
    return NULL;
  }
  srcip = htonl(srcip);
  dstip = htonl(dstip);
  
  map_field((unsigned char *)(&srcip), 4, srcIpMappingTable, &src_ip_count);
  map_field((unsigned char *)(&dstip), 4, dstIpMappingTable, &dst_ip_count);

  return Py_BuildValue("(kk)", ntohl(srcip), ntohl(dstip));
}

static PyObject *ip_map(PyObject *self, PyObject *args) {
  uint32_t ip;
  
  if (! PyArg_ParseTuple(args, "k", &ip)) {
    return NULL;
  }
  ip = htonl(ip);
  map_field((unsigned char *)(&ip), 4, ipMappingTable, &ip_count);
  return Py_BuildValue("k", ntohl(ip));
}

static PyObject *ip_pref_pres_anon(PyObject *self, PyObject *args) {
  uint32_t ip;
  unsigned int anonymized_addr;
  unsigned char p, *an, *raw_addr;
  static int initialized = 0;
  
  if (! PyArg_ParseTuple(args, "k", &ip)) {
    return NULL;
  }
  ip = htonl(ip);

  if (!initialized) {
          PAnonymizer(AES_keys);
          initialized = 1;
  }
  
  raw_addr = (unsigned char *)(&ip);
  
  p = raw_addr[0];
  raw_addr[0] = raw_addr[3];
  raw_addr[3] = p;

  p = raw_addr[1];
  raw_addr[1] = raw_addr[2];
  raw_addr[2] = p;

  anonymized_addr = anonymize(*((unsigned int *)raw_addr));

  an = (unsigned char *)(&anonymized_addr);
  raw_addr[0] = an[3];
  raw_addr[1] = an[2];
  raw_addr[2] = an[1];
  raw_addr[3] = an[0];

  return Py_BuildValue("k", ntohl(ip));
}

static PyObject *ip_pref_pres_map(PyObject *self, PyObject *args) {
  uint32_t ip;
  
  if (! PyArg_ParseTuple(args, "k", &ip)) {
    return NULL;
  }

  /* lookup() operates on _host_ byte order! */
  u_long r_addr = *((u_long *) (&ip));
  addr_propagate.cur_input = r_addr;
  r_addr = lookup(r_addr, &addr_propagate);
  
  return Py_BuildValue("k", r_addr);
}

static PyObject *ip_random(PyObject *self, PyObject *args) {
  uint32_t ip;
  unsigned char hostbits = 8;	/* host bits to mask out */
  int32_t rndval;
  uint32_t rv;
  
  if (! PyArg_ParseTuple(args, "k|b", &ip, &hostbits))
    return NULL;
  
  if ((hostbits <= 0) || (hostbits > 32)) {
    PyErr_SetString(PyExc_ValueError, "bit mask must be 0 < n <= 32");
    return NULL;
  }
  
  rndval = rand32();

  if (hostbits < 32) {
    rndval &= ((1 << hostbits) - 1);
  }
  rv = ip ^ rndval;
  
  return Py_BuildValue("k", rv);
}

static PyObject *ip_truncate(PyObject *self, PyObject *args) {
  uint32_t ip;
  unsigned char hostbits = 8;	/* host bits to mask out */
  
  if (! PyArg_ParseTuple(args, "k|b", &ip, &hostbits))
    return NULL;
  
  if ((hostbits <= 0) || (hostbits > 32)) {
    PyErr_SetString(PyExc_ValueError, "bit mask must be 0 < n <= 32");
    return NULL;
  }
  
  if (hostbits == 32)
    return Py_BuildValue("k", 0);
  
  return Py_BuildValue("k", (ip & (0xffffffff << hostbits)));
}

/*
 * IPv6 support
 */

static PyObject *ip6_aes_hash(PyObject *self, PyObject *args) {
  struct in6_addr ip = IN6ADDR_ANY_INIT;
  
  if (! PyArg_ParseTuple(args, "(IIIIIIII)",
    &(ip.s6_addr16[0]),
    &(ip.s6_addr16[1]),
    &(ip.s6_addr16[2]),
    &(ip.s6_addr16[3]),
    &(ip.s6_addr16[4]),
    &(ip.s6_addr16[5]),
    &(ip.s6_addr16[6]),
    &(ip.s6_addr16[7])
  ))
    return NULL;
  
  aes_hash((unsigned char *)(&ip.s6_addr[0]), 16, 
    (unsigned char *)AES_keys, 8 * sizeof(AES_keys),
    0, NULL);
  
  return Py_BuildValue("(IIIIIIII)", 
    ip.s6_addr16[0],
    ip.s6_addr16[1],
    ip.s6_addr16[2],
    ip.s6_addr16[3],
    ip.s6_addr16[4],
    ip.s6_addr16[5],
    ip.s6_addr16[6],
    ip.s6_addr16[7]
  );
}

static PyObject *ip6_bd_map(PyObject *self, PyObject *args) {
  struct in6_addr sip = IN6ADDR_ANY_INIT;
  struct in6_addr dip = IN6ADDR_ANY_INIT;
  int n;
  
  if (! PyArg_ParseTuple(args, "((IIIIIIII)(IIIIIIII))", 
    &(sip.s6_addr16[0]),
    &(sip.s6_addr16[1]),
    &(sip.s6_addr16[2]),
    &(sip.s6_addr16[3]),
    &(sip.s6_addr16[4]),
    &(sip.s6_addr16[5]),
    &(sip.s6_addr16[6]),
    &(sip.s6_addr16[7]),
    &(dip.s6_addr16[0]),
    &(dip.s6_addr16[1]),
    &(dip.s6_addr16[2]),
    &(dip.s6_addr16[3]),
    &(dip.s6_addr16[4]),
    &(dip.s6_addr16[5]),
    &(dip.s6_addr16[6]),
    &(dip.s6_addr16[7])
  )) {
    return NULL;
  }
  
  for (n = 0; n < 8; n ++) {
    map_field((unsigned char *)(&sip.s6_addr16[n]), 2, srcIp6MappingTable, &src_ip6_count);
    map_field((unsigned char *)(&dip.s6_addr16[n]), 2, dstIp6MappingTable, &dst_ip6_count);
  }

  return Py_BuildValue("((IIIIIIII)(IIIIIIII))",
    sip.s6_addr16[0],
    sip.s6_addr16[1],
    sip.s6_addr16[2],
    sip.s6_addr16[3],
    sip.s6_addr16[4],
    sip.s6_addr16[5],
    sip.s6_addr16[6],
    sip.s6_addr16[7],
    dip.s6_addr16[0],
    dip.s6_addr16[1],
    dip.s6_addr16[2],
    dip.s6_addr16[3],
    dip.s6_addr16[4],
    dip.s6_addr16[5],
    dip.s6_addr16[6],
    dip.s6_addr16[7]
  );
}

static PyObject *ip6_map(PyObject *self, PyObject *args) {
  struct in6_addr ip = IN6ADDR_ANY_INIT;
  int n;
  
  if (! PyArg_ParseTuple(args, "(IIIIIIII)",
    &(ip.s6_addr16[0]),
    &(ip.s6_addr16[1]),
    &(ip.s6_addr16[2]),
    &(ip.s6_addr16[3]),
    &(ip.s6_addr16[4]),
    &(ip.s6_addr16[5]),
    &(ip.s6_addr16[6]),
    &(ip.s6_addr16[7])
  ))
    return NULL;
  
  for (n = 0; n < 8; n ++) {
    map_field((unsigned char *)(&(ip.s6_addr16[n])), 2, ip6MappingTable, &ip6_count);
  }
  
  return Py_BuildValue("(IIIIIIII)", 
    ip.s6_addr16[0],
    ip.s6_addr16[1],
    ip.s6_addr16[2],
    ip.s6_addr16[3],
    ip.s6_addr16[4],
    ip.s6_addr16[5],
    ip.s6_addr16[6],
    ip.s6_addr16[7]
  );
}

/*
 * -===================================================================-
 *  Helpers
 */

uint32_t rand32(void) {
  return ((random() & 0xffff) << 16) | (random() & 0xffff);
}

/* ./. */
