//-- mixmod gui javascript library //-- ysf -----------------------------------
//
//   requires:
//    - jQuery
//    - jQuery ui
//    - jsPlumb
//    - dagre.js
//    - underscore.js
//
//    - #mm-container in html-body

//-- config -------------------------------------------------------------------

var MIXMOD = {};

(function ($, _, jsPlumb, dagre, PNotify) {
  MIXMOD = {
    modules: {}, // has to be provided by mixmod_modules of index.html
    container_id: '#mm-container',
    boxes: {},
    config: {
      connector: ['Bezier', {curviness: 30}],
      endpoints: {
        dragStyle: {fillStyle: '#eee'},
        dropStyle: {fillStyle: '#999'},
        source: {
          anchor: 'Right',
          endpoint: "Dot",
          paintStyle: {width:7, height:7, fillStyle:'#999'}, // you may want this to be the same as target
          connectorStyle: {strokeStyle:"#999", lineWidth: 3},
          isSource: true,
          isTarget: false
        },
        target: {
          endpoint: "Dot",
          anchor: 'Left',
          paintStyle: {width:7, height:7, fillStyle:'#999'}, // you may want this to be the same as dropStyle
          connectorStyle: {strokeStyle:"#999", lineWidth: 3},
          isSource: false,
          isTarget: true
        }
      }
    }
  };

  //-- public methods -----------------------------------------------------------

  MIXMOD.loadData = function (data) {
    var inifile = new INIFile(data);
    drawINIFile(inifile);
  };

  MIXMOD.dumpData = function () {
    var blob = new Blob([makeINIFile()],
                        { type: 'text/json;charset=utf-8' });
    saveAs(blob, "mixmod.conf");
  };

  MIXMOD.initModules = function (mixmod_modules) {
    var menu_entry = _.template($("#template_module_menuentry").html());

    MIXMOD.modules = mixmod_modules.reduce(function(map, obj) {
      map[obj.title] = obj;
      return map;
    }, {});

    _.each(mixmod_modules, function(module, index) {
      var name = module.title;
      $("#" + module.type + "-menu").append(menu_entry({module_name: name}));
      $("#new_" + name).click(function(e){new BoxObject(module);});
    });
  };

  //-- functions ----------------------------------------------------------------

  var _mm_idCounter = 0;
  function uniqueId(prefix) {
    var id = ++_mm_idCounter + '';
    return prefix ? prefix + id : id;
  }

  function BoxObject(module, unique_id, config) {
    this.config = {};
    config = config || {};
    var box = this;

    _.each(config, function (property_value, property_name) {
      if (!(property_name in module.config.items)) {
        warn("Unknown option " + property_name + " for module " + module.title);
      }
    });

    if (module.config !== undefined) {
      _.each(module.config.items, function (property_info, property_name) {
        box.config[property_name] = {
          info: property_info.info,
          type: property_info.type,
          value: config[property_name] || property_info.default
        };
      });
    }

    this.unique_id = unique_id || uniqueId(module.title + '_');
    this.title = this.unique_id;
    this.module_class = module.title;
    this.endpoints = [];
    this.endpoints_data = module.endpoints;
    this.elem = $(this.template({id: this.unique_id,
                                 title: this.title,
                                 subtitle: module.title,
                                 type: module.type}));
    this.incoming_connections = {};
    this.outgoing_connections = {};

    this.generate_box();
    this.add_endpoints();
    this.add_properties();

    MIXMOD.boxes[this.unique_id] = this;
  }

  BoxObject.prototype.template = _.template($("#template_generic_box").html());

  BoxObject.prototype.equals = function(x, y) {
    return _.isEqual(x, y);
  };

  BoxObject.prototype.generate_box = function() {
    $(MIXMOD.container_id).append(this.elem);

    this.elem.find('.closeme').on('click', $.proxy(this.delete_box, this));

    var box = this,
        box_elem = this.elem,
        panel = this.elem.find(".panel-body");

    panel.height(box_elem.height() - 75);

    this.elem.resizable({
      resize: function () {
        panel.height(box_elem.height() - 75);
        jsPlumb.repaintEverything();
      },
      minWidth: 150,
      minHeight: 150
    });

    this.elem.find('.title').editable().on(
      'save', function (e, params) {
        box.title = params.newValue;
      }
    );

    jsPlumb.draggable(this.elem);
  };

  BoxObject.prototype.add_endpoints = function() {
    _.each(this.endpoints_data, function(ep) {
      if(ep in MIXMOD.config.endpoints) {
        var uuid = this.unique_id + '_' + ep;
        this.endpoints.push(
          jsPlumb.addEndpoint(this.unique_id, $.extend(MIXMOD.config.endpoints[ep], {uuid: uuid}))
        );
      }
    }, this);
  };

  BoxObject.prototype.add_properties = function() {
    var property_html = _.template($("#template_box_property").html());

    _.each(this.config,  function(property_info, property_name) {
      var body = $('#'+this.unique_id)
            .find('.panel-body')
            .append(property_html({property: property_name,
                                   value: property_info.value}));
      body.find(".mm-value").last().editable({
        validate: fieldValidator(property_info.type)
      }).on(
        "save",
        function (e, params) {
          property_info.value = params.newValue;
        }
      );
    }, this);
  };

  BoxObject.prototype.delete_box = function(event) {
    jsPlumb.removeAllEndpoints(this.elem);
    this.elem.remove();
    delete MIXMOD.boxes[this.unique_id];
  };

  BoxObject.prototype.connect = function(obj) {
    if("string" === typeof obj)
      obj = MIXMOD.boxes[obj];

    if("undefined" === typeof obj)
      return;

    this.outgoing_connections[obj.unique_id] = obj;
    obj.incoming_connections[this.unique_id] = this;
  };

  BoxObject.prototype.deconnect = function(obj) {
    if("string" === typeof obj)
      obj = MIXMOD.boxes[obj];

    if("undefined" === typeof obj)
      return;

    delete this.outgoing_connections[obj.unique_id];
    delete obj.incoming_connections[this.unique_id];
  };

  BoxObject.prototype.get_parents = function() {
    return this.incoming_connections;
  };

  BoxObject.prototype.get_ancestors = function() {
    var parents = this.get_parents();
    var result = [];
    for(var name in parents) {
      var box = parents[name];
      result.push(box);
      result = result.concat(box.get_ancestors());
    }
    return result;
  };

  BoxObject.prototype.call_endpoints_with = function(fn, param, set_source, set_target) {
    for(var i = 0; i < this.endpoints.length; ++i) {
      var ep = this.endpoints[i];
      if((set_source && ep.isSource) || (set_target && ep.isTarget)) {
        fn.call(ep, param);
      }
    }
  };

  BoxObject.prototype.call_circular_boxes_with = function(fn, param) {
    var boxes = [this].concat(this.get_ancestors());
    for(var id in boxes) {
      fn.call(boxes[id], param, false, true);
    }
  };

  BoxObject.prototype.set_endpoint_enable = function(enabled, set_source, set_target) {
    this.call_endpoints_with(jsPlumb.Endpoint.prototype.setEnabled, enabled, set_source, set_target);
  };

  BoxObject.prototype.set_endpoint_style = function(style, set_source, set_target) {
    this.call_endpoints_with(jsPlumb.Endpoint.prototype.setPaintStyle, style, set_source, set_target);
  };

  // free coke for a better function name push requests
  BoxObject.prototype.set_circular_endpoint_style = function(style) {
    this.call_circular_boxes_with(BoxObject.prototype.set_endpoint_style, style);
  };

  BoxObject.prototype.set_circular_endpoint_enable = function(enabled) {
    this.call_circular_boxes_with(BoxObject.prototype.set_endpoint_enable, enabled);
  };

  function fieldValidator (type) {
    if (type === "int") {
      return function (value) {
        if (!/^[0-9]+$/.test(value)) {
          return "Must be a number";
        } else {
          return null;
        }
      };
    } else if (type === "boolean") {
      return function (value) {
        if (!/^(yes|no|on|off|true|false)$/i.test(value)) {
          return "Must be either yes or no";
        } else {
          return null;
        }
      };
    } else if (type.lastIndexOf("/", 0) === 0) {
      var match = type.match(new RegExp('^/(.*?)/([gimy]*)$'));
      var regex = new RegExp(match[1], match[2]);
      return function (value) {
        if (!regex.test(value)) {
          console.log("|||" + value + "|||");
          return "Must match " + type;
        } else {
          return null;
        }
      };
    } else {
      return function (value) { return null; };
    }
  }

  //-- jsPlumb behavior ---------------------------------------------------------

  function bindEvents() {
    jsPlumb.bind('connection', function(info, originalEvent) {
      MIXMOD.boxes[info.sourceId].connect(info.targetId);
    });

    jsPlumb.bind('connectionDetached', function(info, originalEvent) {
      MIXMOD.boxes[info.sourceId].deconnect(info.targetId);
    });

    jsPlumb.bind('connectionMoved', function(info, originalEvent) {
      // as 'connection'-Event is still fired we only need to remove the old connection
      // and not add one too.
      MIXMOD.boxes[info.sourceId].deconnect(info.targetId);
    });

    jsPlumb.bind('connectionDrag', function(connection) {
      var box = MIXMOD.boxes[connection.sourceId];
      box.set_circular_endpoint_style(MIXMOD.config.endpoints.dragStyle);
      box.set_circular_endpoint_enable(false);
    });

    jsPlumb.bind('connectionDragStop', function(connection) {
      var box = MIXMOD.boxes[connection.sourceId];
      box.set_circular_endpoint_style(MIXMOD.config.endpoints.dropStyle);
      box.set_circular_endpoint_enable(true);
    });
  }

  //-- INI file parser ---

  function INIFile (data) {
    this.sections = this.parseINI(data);
  }

  INIFile.prototype.parseINI = function (data) {
    // we might have sectionless settings
    var currentSection = {name: null, fields: []};
    var sections = [currentSection];

    _.each(data.split(/\r?\n/), function(line) {
      var match;
      if (/^\s*([;#].*)?$/.test(line)) {
        return;
      } else if((match = line.match(/^\[\s*(.*)\s*\]$/))) {
        currentSection = {name: match[1], fields: []};
        sections.push(currentSection);
      } else if((match = line.match(/^([\w\d-_@]+)\s*[=:]\s*(.*)$/))) {
        currentSection.fields.push({name: match[1].trim(), value: match[2].trim()});
      } else if((match = line.match(/^\s*#/))) {
        /* comment intended */
      } else {
        throw new Error("Line '" + line + "' is invalid.");
      }
    });

    return sections;
  };

  INIFile.prototype.get = function (section, option) {
    for (var i = 0; i < this.sections.length; i++) {
      if (this.sections[i].name === section) {
        var fields = this.sections[i].fields;
        for (var j = 0; j < fields.length; j++) {
          if (fields[j].name === option) {
            return fields[j].value;
          }
        }
      }
    }
    return null;
  };

  INIFile.prototype.options = function (section) {
    var options = {};
    for (var i = 0; i < this.sections.length; i++) {
      if (this.sections[i].name === section) {
        var fields = this.sections[i].fields;
        for (var j = 0; j < fields.length; j++) {
          options[fields[j].name] = fields[j].value;
        }
      }
    }
    return options;
  };

  function drawINIFile(inifile) {
    clearGUI();
    createBoxes(inifile);
    autoLayout();
  }

  function createBoxes(inifile) {
    _.each(inifile.sections, function (section) {
      var match;
      if (section.name !== null &&
          (match = section.name.match(/^flowmod:(.*)$/i))) {
        var box_name = match[1],
            class_name = inifile.get(section.name, "class");

        if (!(class_name in MIXMOD.modules)) {
          warn("Unknown module class " + class_name);
          return;
        }

        var module = $.extend({}, MIXMOD.modules[class_name]),
            data_section = inifile.get(section.name, "config"),
            config = data_section ? inifile.options(data_section) :
              {};
        new BoxObject(module, box_name, config);
      }
    });

    _.each(inifile.sections, function (section) {
      var match;
      if (section.name !== null &&
          (match = section.name.match(/^flowmod:(.*)$/i))) {
        var box_name = match[1],
            target_name = box_name + "_target",
            sources = inifile.get(section.name, "src");
        if (sources !== null) {
          _.each(sources.split(" "), function (source) {
            if (!(source in MIXMOD.boxes)) {
              return;
            }
            var source_name = source + "_source";
            jsPlumb.connect({uuids: [source_name, target_name]});
          });
        }
      }
    });
  }

  function clearGUI () {
    jsPlumb.reset();
    $(MIXMOD.container_id).html("");
    MIXMOD.boxes = {};
    bindEvents();
  }

  function autoLayout() {
    var i, elem;
    var g = new dagre.graphlib.Graph();

    g.setGraph({rankdir:'LR'});
    g.setDefaultEdgeLabel(function() { return {}; });

    _.each($('.mm-node'), function(elem) {
      g.setNode(elem.id, {width: elem.clientWidth, height: elem.clientHeight});
    });

    _.each(jsPlumb.getAllConnections(), function(elem) {
      g.setEdge(elem.source.id, elem.target.id);
    });

    dagre.layout(g);

    _.each(g.nodes(), function(elem) {
      $("#" + elem).css("left", g.node(elem).x + "px");
      $("#" + elem).css("top", g.node(elem).y + "px");
    });

    jsPlumb.repaintEverything();
  }

  function makeINIFile () {
    var data = "";
    _.each(MIXMOD.boxes, function (box, name) {
      data += (
        "[FlowMod:" + box.title + "]\n" +
          "class = " + box.module_class + "\n" +
          "config = " + box.title + "\n"
      );
      var sources = [];
      _.each(box.incoming_connections, function (conn, name) {
        sources.push(name);
      });
      if (sources.length > 0) {
        data += "src = " + sources.join(" ") + "\n";
      }
      data += "\n";
      data += "[" + box.title + "]\n";
      _.each(box.config, function (property_info, property_name) {
        if (property_info.value !== null) {
          data += property_name + " = " + property_info.value + "\n";
        }
      });
      data += "\n";
    });
    return data;
  }

  function warn (text) {
    new PNotify({title: "Warning",
                 text: text,
                 type: "error"});
  }

  //-- readyup ------------------------------------------------------------------

  jsPlumb.bind('ready', function() {
    jsPlumb.Defaults.Connector = MIXMOD.config.connector;
    jsPlumb.Defaults.MaxConnections = -1;
    jsPlumb.setContainer($(MIXMOD.container_id));
    bindEvents();
  });

  $.fn.editable.defaults.mode = 'popup';
  $.fn.editable.defaults.toggle = 'dblclick';

})($, _, jsPlumb, dagre, PNotify);
